<?php

namespace App\Console\Commands;

use Illuminate\Console\Command;
use Illuminate\Support\Facades\DB;
use App\Models\User;
use App\Models\Customer;
use App\Models\CustomerAddress;
use App\Traits\OAuth1NetsuiteClientCreator;
use App\Services\Netsuite\NetsuiteCustomersService;
use Illuminate\Support\Facades\Log;

class SyncCustomers extends Command
{
    use OAuth1NetsuiteClientCreator;
    /**
     * The name and signature of the console command.
     *
     * @var string
     */
    protected $signature = 'app:sync-customers';

    /**
     * The console command description.
     *
     * @var string
     */
    protected $description = 'Sync new and existing customers from NetSuite';

    /**
     * Execute the console command.
     */
    public function handle()
    {
        $this->info('Starting synchronization...');
        $start = microtime(true);

        $netsuite = new NetsuiteCustomersService();
        $endpoint = 'https://5874559-sb1.restlets.api.netsuite.com/app/';
        $client = $netsuite->getNetsuiteClient();
        $page = 0;

        try {

            do {
                Log::info("Página $page");
                $customers = $netsuite->fetchCustomersFromNetSuite($client, $page);
                Log::info($customers[0]);
                if (empty($customers)) {
                    break;
                }

                DB::beginTransaction();
                try {
                    $this->processCustomers($customers);
                    DB::commit();
                } catch (\Exception $e) {
                    DB::rollBack();
                    $this->error("Error processing customers on page $page: " . $e->getMessage());
                    Log::error("Error processing customers on page $page: " . $e->getMessage());
                }

                $page++;
            } while (!empty($customers));

            $this->info('Synchronization completed successfully.');

        } catch (\Exception $e) {
            $this->error("Synchronization failed: " . $e->getMessage());
            Log::error("Synchronization failed: " . $e->getMessage());
        }

        $end = microtime(true);
        $time = $end - $start;
        $this->info("----> sync customers duration: {$time}");
    }

    private function processCustomers($customers)
    {
        $this->info("Procesando Clientes...");
        foreach ($customers as $customer) {
            try {
                $this->saveOrUpdateCustomer($customer);
            } catch (\Exception $e) {
                $this->error("Error processing customer: " . $e->getMessage());
                Log::error("Error processing customer: " . $e->getMessage());
            }
        }
        $this->info("Clientes actualizados o creados...");
    }

    private function saveOrUpdateCustomer($customerData)
    {
        //TODO: en sandbox ninguno tiene acceso así que para probar comenté el if-else
        // if($customerData['ACCESO']) {
            $dbUser = User::firstOrCreate(
                ['email' => $customerData['CORREO_ELECTRONICO']],
                ['name' => $customerData['SAT_NOMBRE_REGISTRADO']? $customerData['SAT_NOMBRE_REGISTRADO']: $customerData['NOMBRE']  ]
            );

            $dbCustomer = Customer::updateOrCreate(
                ['rfc' => $customerData['RFC']],
                [
                    'user_id' => $dbUser->id,
                    'company_name' => $customerData['NOMBRE_EMPRESA'],
                    'netsuite_key' => $customerData['ID_CLIENTE'],
                    'rfc' => $customerData['RFC'],
                ]
            );

            $this->updateAddresses($dbCustomer, $customerData['DIRECCIONES']);
        // } else {
        //     Log::info("No tiene acceso: ". $customerData['ACCESO']);
        // }

    }

    private function updateAddresses($dbCustomer, $addresses)
    {
        // Eliminar direcciones antiguas si es necesario
        $dbCustomer->addresses()->delete();

        foreach ($addresses as $address) {
            CustomerAddress::create([
                'customer_id' => $dbCustomer->id,
                'street' => $address['CALLE']." ".$address['NUMERO'],
                'postal_code' => $address['CODIGO_POSTAL'],
                'state' => $address['ESTADO'],
                'country' => $address['PAIS'],
                'number_ext' => $address['NUMERO']
            ]);
        }
    }
}
