<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use App\Models\Product;
use App\Models\ProductStock;
use App\Services\ProductStockService;
use Illuminate\Support\Facades\Log;

class CartController extends Controller
{
    protected $productStockService;

    public function __construct(ProductStockService $productStockService)
    {
        $this->productStockService = $productStockService;
    }

    public function cartList()
    {
        $customer = auth()->user()->customer;

        if (is_null($customer)) {
            session()->flash('message', 'No tienes permiso para realizar esta acción.');
            session()->flash('icon', 'warning');
            return redirect()->back();
        }

        return view('cart');
    }

    public function addToCart(Request $request)
    {
        // TODO: Validar que la cantidad de producto actual en la base de datos ajuste para la cantidad seleccionada por el usuario.

        // Cuando el usuario quiere agregar un producto desde el stock local y nacional
        // En la request viaja quantities que contiene cuánto se agregó en local y en nacional:
        // Y en el siguiente fragmento creamos un uniqueId con la locación (local o nacional) para este producto
        if($request->quantities) {
            $quantities = json_decode($request->quantities, true);
            Log::info($quantities);
            foreach ($quantities as $location => $quantity) {
                if ( (int) $quantity !== 0 && (int) $quantity > 0) {
                    $uniqueId = $request->id . '-' . $location;
                    //Debemos verificar que si hay suficiente stock
                    $locationId = $location == 'Nacional' ? null : auth()->user()->customer->location->id;
                    $product = \Cart::get($uniqueId);
                    $totalQuantity = isset($product) ? ($product->quantity + $quantity) : $quantity;

                    if(ProductStock::hasEnough($request->id, $totalQuantity, $locationId)){
                        \Cart::add([
                            'id' => $uniqueId,
                            'name' => $request->name,
                            'price' => $request->price,
                            'quantity' => $quantity,
                            'attributes' => array(
                                'product_id' => $request->id,
                                'location' => $location,
                                'brand' => $request->brand,
                                'featured' => $request->featured,
                                'weight' => $request->weight,
                                'length' => $request->length,
                                'width' => $request->width,
                                'height' => $request->height,
                                'location_id' => null
                            )
                        ]);
                    } else {
                        session()->flash('message', 'Por el momento ninguna de nuestras sucursales cumple con el stock total solicitado');
                        session()->flash('icon', 'warning');
                        // Borrar Session: por si el usuario ya había hecho alguna cotización de envío pero borra o agrega productos:
                        session()->forget(['local_shipping_data', 'national_shipping_data']);
                        return redirect()->back();
                    }

                }
            }

            session()->flash('message', 'Producto agregado a tu carrito!');
            session()->flash('icon', 'success');

        } else {
            $location = auth()->user()->customer->location->name;
            $uniqueId = $request->id . '-' . $location;
            $product = \Cart::get($uniqueId);
            $totalQuantity = isset($product) ? ($product->quantity + $request->quantity) : $request->quantity;

            if(ProductStock::hasEnough($request->id, $totalQuantity, auth()->user()->customer->location->id)){

                $quantity = $request->quantity;

                \Cart::add([
                    'id' => $uniqueId,
                    'name' => $request->name,
                    'price' => $request->price,
                    'quantity' => $quantity,
                    'attributes' => array(
                        'product_id' => $request->id,
                        'location' => $location,
                        'location_id' => auth()->user()->customer->location->id,
                        'brand' => $request->brand,
                        'featured' => $request->featured,
                        'weight' => $request->weight,
                        'length' => $request->length,
                        'width' => $request->width,
                        'height' => $request->height,
                    )
                ]);

                session()->flash('message', 'Producto agregado a tu carrito!');
                session()->flash('icon', 'success');

            }else{
                session()->flash('message', 'No hay suficiente stock del producto ' . $request->name);
                session()->flash('icon', 'error');
            }
        }

        // Borrar Session: por si el usuario ya había hecho alguna cotización de envío pero borra o agrega productos:
        session()->forget(['local_shipping_data', 'national_shipping_data']);

        return redirect()->back();
    }

    public function updateCart(Request $request)
    {
        //TODO: consultar el stock antes de permitirle actualizar el carrito para comprobar que la cantidad de existe.
        $validated = $request->validate([
            'id' => 'required',
            'quantity' => 'required|numeric',
        ]);

        $product = \Cart::get($request->id);

        if ($product) {

            $locationId = $product->attributes->location_id;
            $id = preg_replace('/\D/', '', $request->id);

            if(ProductStock::hasEnough($id, $request->quantity, $locationId)){

                \Cart::update(
                    $request->id,
                    [
                        'quantity' => [
                            'relative' => false,
                            'value' => $request->quantity
                        ],
                    ]
                );

                $new_subtotal = \Cart::getTotal();

                // Borrar Session: por si el usuario ya había hecho alguna cotización de envío pero borra o agrega productos:
                session()->forget(['local_shipping_data', 'national_shipping_data']);

                $new_subtotal_formatted = view('components.amount-formatter', ['amount' => $new_subtotal])->render();

                return response()->json(['newSubtotalFormatted' => $new_subtotal_formatted]);
            }else
                abort(422, __('No hay suficiente stock del producto'));
        }
    }

    public function removeCart(Request $request)
    {
        \Cart::remove($request->item_id);
        session()->flash('message', 'Elemento eliminado del carrito exitosamente!');
        session()->flash('icon', 'success');

        // Borrar Session: por si el usuario ya había hecho alguna cotización de envío pero borra o agrega productos:
        session()->forget(['local_shipping_data', 'national_shipping_data']);

        return redirect()->back();
    }

    public function clearAllCart()
    {
        \Cart::clear();

        session()->flash('success', 'All Item Cart Clear Successfully !');

        // Borrar Session: por si el usuario ya había hecho alguna cotización de envío pero borra o agrega productos:
        session()->forget(['local_shipping_data', 'national_shipping_data']);

        return redirect()->route('cart.list');
    }

    public function getProductStock(Request $request, Product $product)
    {
        $locationId = auth()->user()->customer->location_id;
        $stock = $this->productStockService->getProductStockForProduct($product, $locationId);
        return response()->json($stock);
    }

}
