<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use App\Mail\PaidOrder;
use App\Mail\PaidOrderAdmin;
use App\Models\Order;
use App\Models\ExchangeRate;
use App\Models\OrderItem;
use App\Models\Shipment;
use Illuminate\Support\Facades\Mail;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Log;
use App\Services\Cart\Factory as CartConditions;
use App\Services\Cart\Rules;
use App\Services\Helpers\Money;
use App\UseCases\Order\Customer\SaveByRequest\Save;
use App\UseCases\Order\Customer\SaveByRequest\Notification;
use App\Services\Teams\Objects\Product as NotificationProduct;
class CheckoutController extends Controller
{
    public function index()
    {
        return view('checkout-details');
    }

    public function getQuoteShipping()
    {
        $locations = [];

        foreach(\Cart::getContent() as $item) {
            $location = $item->attributes->location;
            if (!isset($locations[$location])) {
                $locations[$location] = [
                    'items' => []
                ];
            }
            $locations[$location]['items'][] = $item;
        }

        $customer = auth()->user()->customer;
        $delivery_addresses = $customer->addresses;

        return view('checkout-shipping', compact(
            'locations',
            'customer',
            'delivery_addresses'
        ));
    }

    public function processPayment(Request $request)
    {
        // Decodificar los datos JSON de envío
        $localShippingData = json_decode($request->local_shipping_data, true);
        $nationalShippingData = json_decode($request->national_shipping_data, true);

        // Guardar los datos necesarios en la sesión
        session([
            'local_shipping_data' => $localShippingData,
            'national_shipping_data' => $nationalShippingData,
            'destination_address_id' => $request->destination_customer_address_id
        ]);

        return redirect()->route('checkout.payment');
    }

    public function checkout(Request $request)
    {
        // Obtener los datos de la sesión
        $localShippingData = session('local_shipping_data');
        $nationalShippingData = session('national_shipping_data');

        // Verificar si al menos uno de los datos de envío está disponible
        if (!$localShippingData && !$nationalShippingData) {
            session()->flash('message', 'Faltan datos necesarios para continuar con el pago.');
            session()->flash('icon', 'error');
            return redirect()->route('checkout.get-quote-shipping');
        }

        // Calcular el total con envío en USD
        $totalDetails = $this->getTotalDetails($localShippingData, $nationalShippingData);
        $localShippingUSD = $totalDetails['localShippingUSD'];
        $nationalShippingUSD = $totalDetails['nationalShippingUSD'];
        $totalShippingCostUSD = $totalDetails['totalShippingCostUSD'];
        $totalWithShippingUSD = $totalDetails['totalWithShippingUSD'];

        $customer = auth()->user()->customer;
        $customerId = $customer->id;

        // Generar Folio de Orden:
        $lastFolio = Order::where('customer_id', $customerId)->max('folio');
        $newFolio = $lastFolio ? $lastFolio + 1 : 1;
        $paymentConcept = $this->generatePaymentConcept($newFolio, $customer->company_name);

        // Si los datos no están en la sesión, redirigir a la página anterior
        if (!$paymentConcept) {
            session()->flash('message', 'Faltan datos necesarios para continuar con el pago.');
            session()->flash('icon', 'error');
            return redirect()->route('checkout.get-quote-shipping');
        }

        return view('checkout-payment', [
            'totalDetails' => $totalDetails,
            'localShippingUSD' => $localShippingUSD,
            'nationalShippingUSD' => $nationalShippingUSD,
            'totalShippingCostUSD' => $totalShippingCostUSD,
            'totalWithShippingUSD' => $totalWithShippingUSD,
            'localShippingDetails' => $localShippingData,
            'nationalShippingDetails' => $nationalShippingData,
            'paymentConcept' => $paymentConcept
        ]);
    }


    public function store(Request $request)
    {
        $customer = auth()->user()->customer;

        $rulesForCartConditions = new Rules;

        $rulesForCartConditions->setCarrierLocal($request->input('local_carrier'))
                                ->setCarrierNational($request->input('national_carrier'));

        $handler = new Save($customer, $request->all(), \Cart::getContent(), $rulesForCartConditions);
        $order = $handler->handle();

        if($order){

            \Cart::clearCartConditions();
            \Cart::clear();

            session()->forget(['local_shipping_data', 'national_shipping_data']);
            session()->flash('message', 'Orden guardada exitosamente!');
            session()->flash('icon', 'success');

            return redirect()->route('checkout.complete');
        }else{
            session()->flash('message', 'La transacción no ha podido ser guardada; comunicate con el administrador de la plataforma');
            session()->flash('icon', 'error');

            DB::rollback();
            return redirect()->back()->with('error', 'Error al guardar la orden.');
        }
    }

    public function update(Request $request, Order $order)
    {
        DB::beginTransaction();

        try {
            // Actualiza la orden
            $order->update([
                'payment_concept' => $request->payment_concept,
                'status' => $request->has('payment_submitted') ? 'payment_submitted' : 'pending_payment',
            ]);
            $orderItems = OrderItem::withProduct()
                            ->addSelect('shipments.is_local_shipping')
                            ->join('shipments', 'shipments.id', '=', 'order_items.shipment_id')
                            ->where('order_items.order_id', $order->id)
                            ->get();
            $shipments = Shipment::where('order_id', $order->id)->get();

            Notification::notify(
                Auth::user()->customer,
                $order,
                $orderItems->reduce(function($items, $item){
                    if($item->is_local_shipping) $items[] = new NotificationProduct($item->name, $item->quantity, $item->price, $item->featured);
                    return $items;
                }, []),
                $orderItems->reduce(function($items, $item){
                    if(!$item->is_local_shipping) $items[] = new NotificationProduct($item->name, $item->quantity, $item->price, $item->featured);
                    return $items;
                }, []),
                $shipments->firstWhere('is_local_shipping', 1),
                $shipments->firstWhere('is_local_shipping', 0)
            );
            // Confirmar la transacción
            DB::commit();

            if($request->has('payment_submitted') ) {
                Log::info("Enviando correo.... orden realizada!");
                Mail::to( $order->customer->user->email )->send( new PaidOrder($order) );
                // TODO: enviar al correo que le corresponda.
                Mail::to( 'carlos.hernandez@solar-center.mx' )->send( new PaidOrderAdmin($order) );
            }

            return redirect()->route('checkout.complete');

        } catch (\Exception $e) {
            // Algo salió mal, hacer rollback
            session()->flash('message', 'La transacción no ha podido ser guardada; comunicate con el administrador de la plataforma');
            session()->flash('icon', 'error');
            Log::info($e);

            DB::rollback();
            return redirect()->back()->with('error', 'Error al guardar la orden.');
        }
    }

    public function complete()
    {
        return view('checkout-complete');
    }

    private function generatePaymentConcept($lastFolio, $company_name)
    {
        $folio = sprintf('%04d', $lastFolio);
        return  Auth::user()->customer->netsuite_key .' '. $folio .' '. date('y-m-d');
    }

    private function getTotalDetails($localShippingData = null, $nationalShippingData = null): array
    {
        $rulesForCartConditions = new Rules;
        if($localShippingData)
            $rulesForCartConditions->setCarrierLocal($localShippingData['carrier']);
        if($nationalShippingData)
            $rulesForCartConditions->setCarrierNational($nationalShippingData['carrier']);

        CartConditions::chargeCartCondition($rulesForCartConditions);
        $totalCartUSD = \Cart::getTotal();

        $totalShippingCostUSD = 0;
        $localShippingUSD = 0;
        $nationalShippingUSD = 0;

        if ($localShippingData) {
            $localShippingUSD = Money::convertToDolar($localShippingData['shippingCost']);
            $totalShippingCostUSD = Money::add($totalShippingCostUSD, $localShippingUSD);
        }
        if ($nationalShippingData) {
            $nationalShippingUSD = Money::convertToDolar($nationalShippingData['shippingCost']);
            $totalShippingCostUSD = Money::add($totalShippingCostUSD, $nationalShippingUSD);
        }

        $totalWithShippingUSD = Money::add($totalCartUSD, $totalShippingCostUSD);

        return [
            'localShippingUSD' => $localShippingUSD,
            'nationalShippingUSD' => $nationalShippingUSD,
            'totalShippingCostUSD' => $totalShippingCostUSD,
            'totalWithShippingUSD' => $totalWithShippingUSD
        ];
    }

}
