<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use App\Enums\DeliveryStatusEnum;
use App\Services\TrackingService;
use App\Http\Requests\CustomerContactRequest;
use App\Http\Requests\UpdateCustomerProfileRequest;
use App\Mail\OrderDelivered;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Mail;
use Illuminate\Support\Facades\Hash;
use App\Mail\CustomerContact;
use App\Models\AdditionalChargeApplied;
use App\Models\Customer;
use App\Models\CustomerAddress;
use App\Models\Order;
use App\Models\Shipment;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;

class CustomerAccountController extends Controller
{
    protected $trackingService;

    public function __construct(TrackingService $trackingService)
    {
        $this->trackingService = $trackingService;
    }

    public function profile()
    {
        $customer = auth()->user()->customer;

        $orders = Order::all();
        return view('customer.account-profile', [
            'customer' => $customer,
            'orders' => $orders,
            'delivery_addresses' => $customer->addresses
        ]);
    }

    public function profileUpdate(UpdateCustomerProfileRequest $request, Customer $customer)
    {
        // Actualizar nombre y correo electrónico
        $customer->user->name = $request->name;
        $customer->user->email = $request->email;
        $customer->user->save();
        $customer->save();
        // Actualizar contraseña si se proporcionó una nueva
        if (!empty($request->password)) {
            $customer->user->password = Hash::make($request->password);
            $customer->save();
            $customer->user->save();
        }
        //Actualizar direccion predeterminada
        $customer = auth()->user()->customer;
        CustomerAddress::where('customer_id', $customer->id)
            ->where('id', $request->input('default_address'))
            ->limit(1)
            ->update(['is_default' => true]);

        session()->flash('message', 'Perfil actualizado con éxito');
        session()->flash('icon', 'success');

        return redirect()->back();
    }

    public function locationUpdate(Request $request)
    {
        $customer = Auth::user()->customer;
        $customer->location_id = $request->location_id;
        $customer->save();
        return redirect()->back();
    }

    public function orders(Request $request)
    {
        $customer = Auth::user()->customer;

        $query = Order::select(
                    'orders.id',
                    'orders.customer_id',
                    'orders.folio',
                    'orders.total',
                    'orders.payment_concept',
                    'orders.status',
                    'orders.invoice',
                    'orders.updated_at',
                    'orders.deleted_at',
                    'orders.created_at',
                    DB::raw('cancellation_requests.created_at AS cancellationRequest'))
                ->leftJoin('cancellation_requests', 'cancellation_requests.order_id', '=', 'orders.id')
                ->orderBy('orders.id', 'DESC');

        // Aplica filtros
        if ($request->filled('status')) {
            $query->where('orders.status', $request->input('status'));
        }

        if ($request->filled('created_at')) {
            $date = $request->input('created_at');
            $query->whereDate('orders.created_at', $date);
        }

        // Pagina el resultado
        $orders = $query->paginate(5);

        // Para generar el concepto de pago:
        $last_order = Order::where('customer_id', $customer->id)->orderBy('created_at','DESC')->first();
        $last_order_id = !is_null($last_order) ? $last_order->id : 1;
        $ordersId = $orders->pluck('id');
        $shipments = Shipment::whereIn('order_id', $ordersId)->get();
        $additionalChargesApplied = AdditionalChargeApplied::full()->whereIn('additional_charges_applied.order_id', $ordersId)->get();

        foreach ($orders as $order) {

            $order->payment_concept = $this->generatePaymentConcept($last_order_id, $customer->company_name);
            $order->additionalChargesApplied = $additionalChargesApplied->filter(fn($additionalChargeApplied) => $additionalChargeApplied->order_id == $order->id);
            $order->shipments = $shipments->filter(fn(Shipment $shipment) => $shipment->order_id == $order->id);
            $order->tracking_info = $shipments->reduce(function($items, $shipment) use ($order) {
                if($shipment->order_id == $order->id){

                    $items[] = (object)[
                        'tracking' => $shipment->master_tracking_number ? $shipment->tracking() : null,
                        'shipment' => $shipment
                    ];
                }

                return $items;
            }, []);
        }

        return view('customer.account-orders', [
            'customer' => $customer,
            'orders' => $orders,
            'status' => $request->has('status') ? $request->status : null,
            'delivery_status' => $request->has('delivery_status') ? $request->delivery_status : null,
            'created_at' => $request->has('created_at') ? $request->created_at : null,
            'additionalChargesApplied' => $additionalChargesApplied
        ]);
    }

    public function contact()
    {
        return view('customer.contact');
    }

    public function sendCustomerContact(CustomerContactRequest $request)
    {
        $name = $request->name;
        $email = $request->email;
        $phone = $request->phone;
        $subject = $request->subject;
        $message_body = $request->message;

        try {
            Mail::to( $email )->send( new CustomerContact( $name, $email, $phone, $subject, $message_body ) );
            session()->flash('message', 'Tu mensaje ha sido enviado con éxito!');
            session()->flash('icon', 'success');
        } catch (\Exception $e) {
            session()->flash('message', 'Hubo un error al enviar tu mensaje. Por favor, intenta de nuevo más tarde.');
            session()->flash('icon', 'error');
            Log::error('Error al encolar correo: ' . $e->getMessage());
        }

        return redirect()->back();
    }

    private function generatePaymentConcept($last_order_id, $company_name)
    {
        $folio = sprintf('%04d', $last_order_id);

        // Divide el nombre de la empresa en palabras y toma la primera palabra
        $company_words = explode(' ', $company_name);
        // $first_word_of_company_name = $company_words[0];

        // return  Auth::user()->customer->netsuite_key .' '. $folio .' '. $first_word_of_company_name .' '. date('y-m-d');
        return  Auth::user()->customer->netsuite_key .' '. $folio .' '. date('y-m-d');
    }

}
