<?php

namespace App\Http\Controllers\CustomerSupport;

use App\Http\Controllers\Controller;
use Illuminate\View\View;
use App\Http\Requests\UpdateOrderStatusRequest;
use App\Http\Requests\UpdateOrderInvoiceRequest;
use App\Models\Order;
use App\Enums\DeliveryServicesEnum;
use App\Http\Resources\CustomerSupport\Order as CustomerSupportOrder;
use App\Models\OrderedItemInPackagesSent;
use App\Models\OrderItem;
use App\Models\PackageSent;
use App\Services\TrackingService;
use App\Models\Shipment;
use App\UseCases\PackagesSent\Viewer\Shipments as ViewShipments;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Storage;
use App\UseCases\PackagesSent\Viewer\ShipmentsPropose;

class OrderController extends Controller
{
    protected $trackingService;

    public function __construct(TrackingService $trackingService)
    {
        $this->trackingService = $trackingService;
    }

    public function index(): View
    {
        $orders = Order::select(
                    'orders.id',
                    'orders.customer_id',
                    'orders.folio',
                    'orders.total',
                    'orders.payment_concept',
                    'orders.status',
                    'orders.invoice',
                    'orders.method',
                    'orders.created_at',
                    'orders.updated_at',
                    'orders.deleted_at',
                    'customers.company_name',
                    DB::raw('cancellation_requests.created_at AS cancellationRequest')
                    )
                    ->orderBy('created_at', 'DESC')
                    ->join('customers', 'customers.id', '=', 'orders.customer_id')
                    ->leftJoin('cancellation_requests', 'cancellation_requests.order_id', '=', 'orders.id')
                    ->get();

        return view('customer_support.orders.index', ['orders' => CustomerSupportOrder::collection($orders)]);
    }

    public function edit($id): View
    {
        $order = Order::select(
                'orders.id',
                'orders.customer_id',
                'orders.folio',
                'orders.total',
                'orders.payment_concept',
                'orders.status',
                'orders.invoice',
                'orders.method',
                'orders.created_at',
                'orders.updated_at',
                'orders.deleted_at',
                DB::raw('cancellation_requests.created_at AS cancellationRequest')
            )
            ->orderBy('created_at', 'DESC')
            ->leftJoin('cancellation_requests', 'cancellation_requests.order_id', '=', 'orders.id')
            ->findOrFail($id);
        $courierOptions = DeliveryServicesEnum::getCourierOptions();
        $shipments = Shipment::fullQuery()->where('order_id', $order->id)->get();
        $trackings = $shipments->map(fn(Shipment $shipment) => [
            'tracking' => $shipment->master_tracking_number ? $shipment->tracking() : null,
            'shipment_id' => $shipment->id
        ]);
        $items = OrderItem::withProduct()
                    ->where('order_items.order_id', $order->id)
                    ->get();
        $itemsGroupedByShipping = $items->groupBy('shipment_id');
        $shipmentIds = $shipments->pluck('id');
        $packagesSent = PackageSent::select(
                            'packages_sent.id',
                            'packages_sent.shipping_box_id',
                            'packages_sent.shipment_id',
                            'packages_sent.total_items',
                            'packages_sent.tracking_id',
                            DB::raw('shipping_boxes.name AS shipping_box_name'),
                            DB::raw('shipping_boxes.outer_length AS shipping_box_outer_length'),
                            DB::raw('shipping_boxes.outer_width AS shipping_box_outer_width'),
                            DB::raw('shipping_boxes.outer_height AS shipping_box_outer_height'),
                            DB::raw('shipments.is_local_shipping'))
                        ->leftJoin('shipping_boxes', 'shipping_boxes.id', '=', 'packages_sent.shipping_box_id')
                        ->join('shipments', 'shipments.id', '=', 'packages_sent.shipment_id')
                        ->whereIn('packages_sent.shipment_id', $shipmentIds)
                        ->get()
                        ->groupBy('shipment_id')
                        ->all();
        $orderedItemsInPackages = OrderedItemInPackagesSent::select(
                                    'ordered_items_in_packages_sent.id',
                                    'ordered_items_in_packages_sent.packages_sent_id',
                                    'ordered_items_in_packages_sent.order_item_id',
                                    'ordered_items_in_packages_sent.x',
                                    'ordered_items_in_packages_sent.y',
                                    'ordered_items_in_packages_sent.z',
                                    'ordered_items_in_packages_sent.length',
                                    'ordered_items_in_packages_sent.width',
                                    'ordered_items_in_packages_sent.depth')
                                    ->get()
                                    ->groupBy('packages_sent_id')
                                    ->all();

        $packagesView = new ViewShipments($packagesSent, $orderedItemsInPackages, $shipmentIds, $items);
        $proposePackagesView = new ShipmentsPropose(
                                    $shipments->reduce(function($acc, $shipment){
                                        if(!$shipment->master_tracking_number) $acc[] = $shipment->id;
                                        return $acc;
                                    }, []),
                                    $items
                                );
        return view('customer_support.orders.edit', [
            'order' => $order,
            'courierOptions' => $courierOptions,
            'shipments' => $shipments,
            'trackings' => $trackings,
            'can_cancel' => $order->canCancel(),
            'can_update' => $order->canUpdate(),
            'additionalCharges' => $order->additionalCharges(),
            'itemsGroupedByShipping' => $itemsGroupedByShipping,
            'packagesView' => $packagesView->handle(),
            'proposePackagesView' => $proposePackagesView->handle()
        ]);

    }

    public function updateStatus(UpdateOrderStatusRequest $request, Order $order)
    {
        $order->status = $request->status;
        $order->save();

        session()->flash('message', 'Status de la orden actualizado exitosamente');
        session()->flash('icon', 'success');

        return redirect()->back();
    }

    public function updateInvoice(UpdateOrderInvoiceRequest $request, Order $order)
    {

        if ($request->hasFile('invoice')) {
            // Verificar si ya existe un archivo de factura y eliminarlo
            if ($order->invoice && Storage::disk('public')->exists($order->invoice)) {
                Storage::disk('public')->delete($order->invoice);
            }

            $order->invoice = $request->file('invoice')->store('invoice', 'public');
            $order->save();
            session()->flash('message', 'La factura de la orden se actualizó exitosamente');
            session()->flash('icon', 'success');
        }

        return redirect()->route('admin.orders.index');
    }
    public function cancel(Request $req, int $id){
        $validated = $req->validate([
            'reason' => 'required',
        ]);

        $order = Order::findOrFail($id);

        if($order->canCancel()){

            if($order->cancel(Auth::id(), $req->input('reason'))) return response()->json(['message' => __('alerts.admin-canceled-order')]);

            return abort(500, __('alerts.error'));
        }

        abort(422, __('alerts.admin-can-not-cancel-order'));
    }
}
