<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use App\Models\Product;
use App\Models\Category;
use Illuminate\Support\Facades\Storage;
use App\Http\Requests\CustomerContactRequest;
use App\Http\Requests\IntegratorContactRequest;
use App\Http\Resources\Welcome\Products;
use Illuminate\Support\Facades\Mail;
use App\Mail\CustomerContact;
use App\Models\HeroSlider;
use App\Models\BannerGroup;
use App\Services\ProductStockService;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;

class HomeController extends Controller
{
    protected $productStockService;
    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct(ProductStockService $productStockService)
    {
        $this->productStockService = $productStockService;
    }

    /**
     * Show the application dashboard.
     *
     * @return \Illuminate\Contracts\Support\Renderable
     */
    public function index(Request $request)
    {
        $locationId = auth()->user()->customer->location_id ?? null;

        $categoryId = $request->input('category_id');
        $brand = $request->input('brand');

        $query = Product::queryStore($locationId, $categoryId, $brand)
                    ->orderBy('id', 'DESC');

        $products = $query->paginate(12);
        $sliders = HeroSlider::orderBy('id', 'DESC')->get();
        $banners = BannerGroup::orderBy('id', 'DESC')->get();

        return view('index', [
            'products' => Product::formatForIndex($products),
            'sliders' => $sliders,
            'banners' => $banners
        ]);

    }

    public function productFilter(Request $request)
    {
        $locationId = auth()->user()->customer->location_id ?? null;
        $categoryId = $request->input('category_id');
        $brand = $request->input('brand');

        $query = Product::queryStore($locationId, $categoryId, $brand)
                    ->orderBy('id', 'DESC');

        // Ejecuta la consulta y obtiene los resultados
        $products = $query->paginate(12);

        return view('products.index', ['products' => Product::formatForIndex($products)]);
    }

    public function productByFilter(Request $request, Product $product)
    {
        $product = Product::select(
                            'products.id',
                            'products.name',
                            'products.featured',
                            'products.length',
                            'products.width',
                            'products.height',
                            'products.weight',
                            'products.description',
                            'products.brand',
                            DB::raw('product_prices.price AS defaultPrice'),
                            DB::raw('categories.name AS category'))
                        ->join('product_prices', 'product_prices.product_id', '=', 'products.id')
                        ->leftJoin('categories', 'categories.id', '=', 'products.category_id')
                        ->where('product_prices.level', 'Lista 1')
                        ->findOrFail($product->id);
        // $product->featured_url = Storage::url($product->featured);
        $product->data_sheet_url = $product->data_sheet ? Storage::url($product->data_sheet) : null;

        // Loads stocks:
        if (auth()->check()) {
            $locationId = auth()->user()->customer->location_id ?? null;
            $productStockService = new ProductStockService();
            $stock = $productStockService->getProductStockForProduct($product, $locationId);
            $product->localStock = $stock['localStock'];
            $product->nationalStock = $stock['nationalStock'];
        }

        // Obtener productos de la misma marca y categoría excluyendo el producto actual
        $relatedByBrand = Product::whereNotNull('brand')
            ->where('brand', $product->brand)
            ->where('id', '!=', $product->id)
            ->take(5)
            ->get();

        $relatedByCategory = Product::whereNotNull('category_id')
            ->where('category_id', $product->category_id)
            ->where('id', '!=', $product->id)
            ->take(5)
            ->get();

        // Combina las colecciones y elimina duplicados
        $relatedProducts = $relatedByBrand->merge($relatedByCategory)->unique('id');
        // Mezcla los productos aleatoriamente
        $relatedProducts = $relatedProducts->shuffle();

        return view('product', compact('product', 'relatedProducts'));
    }

    public function productsByCategory(Request $request, Category $category)
    {
        $products = Product::where('category_id', $category->id)->get();

        return view('category', compact('products'));

    }

    public function searchProducts(Request $request)
    {
        $searchTerm = $request->search_term;

        $products = Product::query();
        $products->with('category');

        if ($searchTerm) {
            $products = $products->where(function ($query) use ($searchTerm) {
                $query->where('name', 'LIKE', "%{$searchTerm}%")
                    ->orWhere('brand', 'LIKE', "%{$searchTerm}%")
                    ->orWhereHas('category', function ($subQuery) use ($searchTerm) {
                        $subQuery->where('name', 'LIKE', "%{$searchTerm}%");
                    });
            });
        } else {
            $products = $products->orderBy('id', 'DESC');
        }

        $products = $products->get()->map(function ($product) {
            $product->featured_url = Storage::url($product->featured);
            $product->data_sheet_url = Storage::url($product->data_sheet);
            return $product;
        });

        return response()->json(['products' => $products]);
    }

    public function filterByCategory(Request $request)
    {
        return view('category', compact('products'));
    }

    public function aboutUs()
    {
        return view('about_us.index');
    }

    public function branches()
    {
        return view('about_us.branches');
    }

    public function contact()
    {
        return view('about_us.contact');
    }

    public function sendContact(CustomerContactRequest $request)
    {
        $name = $request->name;
        $email = $request->email;
        $phone = $request->phone;
        $subject = $request->subject;
        $message_body = $request->message;

        try {
            Mail::to( $email )->send( new CustomerContact( $name, $email, $phone, $subject, $message_body ) );
            session()->flash('message', 'Tu mensaje ha sido enviado con éxito!');
            session()->flash('icon', 'success');
        } catch (\Exception $e) {
            session()->flash('message', 'Hubo un error al enviar tu mensaje. Por favor, intenta de nuevo más tarde.');
            session()->flash('icon', 'error');
            Log::error('Error al encolar correo: ' . $e->getMessage());
        }

        return redirect()->back();
    }

    public function services()
    {
        return view('services.index');
    }

    public function offer()
    {
        return view('services.offer');//2
    }

    public function integrators()
    {
        return view('integrators.index');
    }

    public function sendIntegratorContact(IntegratorContactRequest $request)
    {
        $name = $request->name;
        $email = $request->email;
        $phone = $request->phone;
        $state = $request->state;
        $company = $request->company;
        $rfc = $request->rfc;
        $message_body = $request->message;

        try {
            Mail::to( $email )->send( new CustomerContact( $name, $email, $phone, $state, $company, $rfc, $message_body ) );
            session()->flash('message', 'Tu mensaje ha sido enviado con éxito!');
            session()->flash('icon', 'success');
        } catch (\Exception $e) {
            session()->flash('message', 'Hubo un error al enviar tu mensaje. Por favor, intenta de nuevo más tarde.');
            session()->flash('icon', 'error');
            Log::error('Error al encolar correo: ' . $e->getMessage());
        }

        return redirect()->back();
    }

    public function howToBuy()
    {
        return view('faqs.como-comprar');
    }

    public function questions()
    {
        return view('faqs.questions');
    }

    public function generalPolicies()
    {
        return view('information.general-policies');
    }

    public function warrantyPolicies()
    {
        return view('information.warranty-policies');
    }

    public function noticePrivacy()
    {
        return view('information.notice-privacy');
    }
}
