<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use App\Models\Product;
use App\Models\ShippingBox;
use App\Services\Helpers\Money;
use App\Services\Shipments\Addresess;
use App\Services\Shipments\Address;
use App\Services\Shipments\Product as ShipmentsProduct;
use App\Services\Shipments\ShippingQuote;
use App\Services\ShippingServiceFactory;
use Darryldecode\Cart\CartCollection;
use Illuminate\Support\Facades\Log;
use App\UseCases\Shipment\QuotationInCache;
use Illuminate\Support\Facades\Cache;
use App\Services\SelfPickup\Service as SelfPickupService;
use App\Services\Shipments\PackagedShipping;

class ShippingController extends Controller
{

    protected $shippingServiceFactory;

    public function __construct(ShippingServiceFactory $factory)
    {
        $this->shippingServiceFactory = $factory;
    }

    public function quoter(Request $request)
    {
        $carriers = ShippingServiceFactory::companies();
        $response = ['quotes' => []];
        $response['selfPickupCost'] = SelfPickupService::defaultCost();
        $quotationInCache = new QuotationInCache(Cache::getStore());

        $destination = auth()->user()->customer->addresses->find($request->address);

        $location = $request->location === 'Nacional' ? 'Nacional' : 'Local';
        $response['quoteType'] = $location;

        $products = $this->getProductsCartByLocation($location);


        $productsIds = $products->reduce(function($ids, $item){
                        $ids[] = $item->attributes->product_id;
                        return $ids;
                    }, []);
        $response['productsIds'] = $productsIds;

        $shipment = $this->shipment(
                        new Address($destination->id, $destination->postal_code, $destination->neighborhood),
                        $location,
                        $products,
                        $productsIds
                    );
        $shipment->calculate();
        $response['shipment'] = $shipment;

        foreach ($carriers as $carrier) {
            try {
                $shippingService = $this->shippingServiceFactory->make($carrier);

                $quoteResponse = $shippingService->getQuote($shipment);

                $response['quotes'][$carrier] = $quoteResponse;
                //Guardar el costo de la cotizacion en la cache para ser usada si el cliente decide comprar sus productos
                //Todas las cotizaciones vienen en pesos
                $quotationInCache->put(
                    $request->location === 'Nacional' ? 'national' : 'local',
                    $carrier,
                    $quoteResponse['Quotation'][0]['Service'][0]['ServiceCode'],
                    Money::convertToDolar($quoteResponse['Quotation'][0]['Service'][0]['TotalAmount'])
                );

            } catch (\Throwable $e) {
                Log::error("Error al obtener cotización de {$carrier}: " . $e->getMessage());
                Log::error("Error al obtener cotización de {$carrier}: " . $e->getTraceAsString());
                $response['quotes'][$carrier] = [
                    'error' => "No se pudo obtener cotización de {$carrier}."
                ];
            }
        }

        return response()->json($response);
    }
    private function shipment(Address $destination, $location, CartCollection $products, array $productsIds): ShippingQuote{
        $productsDb = Product::select(
                        'products.id',
                        'products.name',
                        'products.length',
                        'products.width',
                        'products.height',
                        'products.weight',
                        'product_prices.price',
                        'products.sat_clave_producto_servicio',
                            'products.sat_clave_unidad_peso')
                        ->join('product_prices', 'product_prices.product_id', '=', 'products.id')
                        ->whereIn('products.id', $productsIds)
                        ->where('product_prices.level', 'Lista 1')
                        ->get();
        $addresessWithStock = $this->getAddresessWithStock($location, $products, $productsIds);
        $productsForPackaging = [];

        foreach ($products as $product) {
            $productDb = $productsDb->firstWhere('id', $product->attributes->product_id);
            $productsForPackaging[]  = new ShipmentsProduct(
                $productDb->id,
                $productDb->name,
                $productDb->length,
                $productDb->width,
                $productDb->height,
                $productDb->weight,
                $product->quantity,
                $productDb->price,
                $productDb->sat_clave_producto_servicio,
                $productDb->sat_clave_unidad_peso
            );
        }

        $shipment = new ShippingQuote($destination, $addresessWithStock, ShippingBox::boxList(), new PackagedShipping($productsForPackaging));


        return $shipment;

    }
    private function getAddresessWithStock($location, CartCollection $products, array $productsIds){

        $stock = $this->getProductsStockInLocations($location, $productsIds);
        $addresses = new Addresess;
        //Recorremos todas las ubicaciones y sus stocks
        foreach ($stock as $locationId => $stockData) {

            $hasStock = $products->reduce(function($hasStock, $product) use($stockData){

                if(!key_exists($product->attributes->product_id, $stockData['products_stock']))
                    $hasStock = false;
                else if($stockData['products_stock'][$product->attributes->product_id] < $product->quantity) $hasStock = false;

                return $hasStock;
            }, true);

            if($hasStock)
                $addresses->add(
                    new Address(
                        $locationId,
                        $stockData['postal_code'],
                        $stockData['neighborhood'],
                        $stockData['street'],
                        $stockData['number_ext'],
                        $stockData['city'],
                        $stockData['state'])
                );

        }
        /** Si no hay ningun codigo postal se entiende que alguno de los productos (o todos) no tienen el stock suficiente*/
        if(count($addresses) == 0) throw new \Exception(__('shipment.insufficient-stock'));

        return $addresses;

    }

    private function getProductsStockInLocations($location, $productIds){
        /**Obtenemos el stock ordenado por id de producto y id de la ubicacion */
        $stock =  Product::select(
                        'products.id',
                        'product_stocks.location_id',
                        'locations.postal_code',
                        'locations.neighborhood',
                        'locations.street',
                        'locations.number_ext',
                        'locations.city',
                        'locations.state',
                        'product_stocks.quantity_available')
                    ->leftJoin('product_stocks', 'product_stocks.product_id', '=', 'products.id')
                    ->join('locations', 'locations.id', '=', 'product_stocks.location_id')
                    ->where('product_stocks.quantity_available', '>', 0)
                    ->whereIn('products.id', $productIds)
                    ->whereNotNull('product_stocks.quantity_available')
                    ->whereNotNull('locations.postal_code')
                    ->when($location == 'Local', function($query){
                        /**Si el carrito es de el stock local buscamos el almacen donde esta ubicado el cliente */
                        return $query->where('product_stocks.location_id', auth()->user()->customer?->location->id);

                    })
                    ->orderBy('products.id', 'ASC')
                    ->orderBy('product_stocks.location_id', 'ASC')
                    ->orderBy('locations.postal_code', 'ASC')
                    ->orderBy('product_stocks.quantity_available', 'ASC')
                    ->get();
        /** Ahora retornamos el stock agrupado por ubicacion  */
        return $stock->reduce(function($stockOrdered, $item){

            if(!key_exists($item->location_id, $stockOrdered)){
                $stockOrdered[$item->location_id] = [
                    'postal_code' => $item->postal_code,
                    'neighborhood' => $item->neighborhood,
                    'street' => $item->street,
                    'number_ext' => $item->number_ext,
                    'city' => $item->city,
                    'state' => $item->state,
                    'products_stock' => []
                ];
            }

            $stockOrdered[$item->location_id]['products_stock'][$item->id] = $item->quantity_available;

            return $stockOrdered;
        }, []);
    }

    private function getProductsCartByLocation($location){
        $location = strtolower($location);

        if($location == 'nacional')
            return \Cart::getContent()->filter(function($item){
                $condition = strpos($item->id, 'Nacional');

                return $condition !== false ;
            });

        return \Cart::getContent()->filter(function($item){
            $condition = strpos($item->id, 'Nacional');

            return $condition === false ;
        });
    }

}
