<?php

namespace App\Models;

use App\Enums\OrderStatusEnum;
use App\Models\CancellationRequest as ModelsCancellationRequest;
use App\UseCases\Order\Cancel;
use App\UseCases\Order\Customer\CancellationRequest;
use Illuminate\Database\Eloquent\Factories\HasFactory;
use Illuminate\Database\Eloquent\SoftDeletes;
use Illuminate\Database\Eloquent\Model;

class Order extends Model
{
    use HasFactory, SoftDeletes;

    public $additionalCharges = [];

    protected $fillable = [
        'customer_id',
        'folio',
        'total',
        'payment_concept',
        'status',
        'courier_code',
        'delivery_status',
        'delivery_event',
        'estimated_delivery_date',
    ];
    public function customer()
    {
        return $this->belongsTo(Customer::class);
    }

    public function items()
    {
        return $this->hasMany(OrderItem::class);
    }

    public function shipments()
    {
        return $this->hasMany(Shipment::class);
    }

    public function folio()
    {
        $folio = sprintf('%04d', $this->id);
        return $folio;
    }
    /**
     * Solo actualizamos la orden si ya todos los envios tienen ya una guia
     *
     * @param integer $order_id
     * @return void
     */
    public static function updateStatusWithGuide(int $order_id){
        $shipmentWithOutTrack = Shipment::where('order_id', $order_id)->whereNull('master_tracking_number')->first();

        if(!$shipmentWithOutTrack)
            self::where('id', $order_id)->update(['status' => OrderStatusEnum::WITH_GUIDE]);
    }
    public static function statusTextColor(string $status){
        switch ($status) {
            case OrderStatusEnum::PAYMENT_SUBMITTED:
                return 'text-bg-warning';
            case OrderStatusEnum::PENDING_PAYMENT:
                return 'text-bg-info';
            case OrderStatusEnum::APPROVED:
                return 'text-bg-success';
            case OrderStatusEnum::CANCELLED:
                return 'text-bg-light';
            default:
                return 'text-bg-info';
        }
    }
    /**
     * Saber si todos los envios de una orden estan cancelados o no tienen guia
     *
     * @param integer $orderId
     * @return boolean
     */
    public static function hasAllShipmentsInactives(int $orderId){
        $shipmentNotCanceled = Shipment::select('shipments.id', 'shipments_canceled.created_at')
                                ->leftJoin('shipments_canceled', 'shipments_canceled.shipment_id', '=', 'shipments.id')
                                ->where(function($query){
                                    $query->whereNull('shipments_canceled.created_at')
                                            ->whereNotNull('shipments.master_tracking_number');
                                })
                                ->where('shipments.order_id', $orderId)
                                ->first();
        return $shipmentNotCanceled === null ? true  : false;
    }
    /**
     * Cancelar una orden, para poder hacerlo todos los envios deben estar cancelados
     *
     * @return boolean
     */
    public function canCancel(){
        return Cancel::can($this);
    }

    public function cancel(int $useId, string $reason){
        $handler = new Cancel(
                        $useId,
                        $this,
                        User::fullCustomer()
                            ->where('customers.id', $this->customer_id)
                            ->firstOrFail(),
                        $reason
                    );

        return $handler->handle();
    }
    public function canCancellationRequest(){
        return CancellationRequest::canByStatus($this);
    }
    public function canUpdate(){
        return match ($this->status) {
            OrderStatusEnum::CANCELLED => false,
            default => true,
        };
    }
    public function additionalCharges(){

        if($this->additionalCharges) return $this->additionalCharges;

        $this->additionalCharges = AdditionalChargeApplied::byOrderId($this->id)->get();

        return  $this->additionalCharges;
    }
    public function hasPendingCancellationRequest(){
        $order = ModelsCancellationRequest::where('order_id', $this->id)->first();

        return isset($order);
    }
}
