<?php

namespace App\Models;

use Illuminate\Database\Eloquent\Factories\HasFactory;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\SoftDeletes;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Storage;

class Product extends Model
{
    use HasFactory, SoftDeletes;

    protected $guarded = [];
    /**
     * The "booted" method of the model.
     */
    protected static function booted(): void
    {
        static::addGlobalScope('valid_dimensiones', function (Builder $builder) {
            if(env('APP_PRODUCTS_VALID_DIMENSIONES'))
                $builder->whereIn('products.id', [208,210,31, 6]);
        });
    }
    /**
     * Local Scope
     */
    public function scopeValidForEcommerce(Builder $query): void
    {
        $query->whereNotNull('featured')
                ->whereNotNull('weight')
                ->whereNotNull('length')
                ->whereNotNull('width')
                ->whereNotNull('height');
    }

    public function category()
    {
        return $this->belongsTo(Category::class, 'category_id');
    }

    /**
     * ------ Gestión del precio ------
    */
    public function prices()
    {
        return $this->hasMany(ProductPrice::class);
    }
    /**
     * ------ Gestión del stock ------
    */
    public function stocks()
    {
        return $this->hasMany(ProductStock::class);
    }

    public function getTotalAvailableQuantityAttribute()
    {
        return $this->stocks->sum('quantity_available');
    }
    /**
     * Consulta para verificar el stock disponible
     *
     * @return mixed
     */
    public static function queryExistsStock(){
        return self::select(
                    'products.id',
                    'product_stocks.location_id',
                    'product_stocks.quantity_available',
                    'locations.postal_code')
                ->leftJoin('product_stocks', 'product_stocks.product_id', '=', 'products.id')
                ->join('locations', 'locations.id', '=', 'product_stocks.location_id')
                ->where('product_stocks.quantity_available', '>', 0)
                ->whereNotNull('product_stocks.quantity_available')
                ->orderBy('products.id', 'ASC')
                ->orderBy('product_stocks.location_id', 'ASC')
                ->orderBy('locations.postal_code', 'ASC')
                ->orderBy('product_stocks.quantity_available', 'ASC');
    }
    /**
     * Consulta para obtener los siguientes datos:
     *
     * - Productos disponibles
     * - Precio principal de cada producto
     * - Stock nacional de cada producto
     * - Stock específico en la sucursal asignada (Si se proporciona un ID de sucursal)
     * - ID de la sucursal asignada (Si se proporciona un ID de sucursal)
     * - Nombre de la sucursal asignada en Netsuite (Si se proporciona un ID de sucursal)
     * - Tambien se puede filtrar por categoria o marca
     * @param integer|null $location_id
     * @param integer|null $categoryId
     * @param string|null $brand
     * @return object
     */
    public static function queryStore(int $location_id = null, int $categoryId = null, string $brand = null){

        $query = self::select(
                    'products.id',
                    'products.name',
                    'products.length',
                    'products.width',
                    'products.height',
                    'products.weight',
                    'products.featured',
                    'products.data_sheet',
                    'products.category_id',
                    'products.brand',
                    'products.description',
                    DB::raw('product_prices.price AS defaultPrice'),
                    DB::raw('SUM(product_stocks.quantity_available) AS nationalStock'),
                    DB::raw('categories.name AS category'))
                ->join('product_prices', 'product_prices.product_id', '=', 'products.id')
                ->leftJoin('product_stocks', 'product_stocks.product_id', '=', 'products.id')
                ->leftJoin('categories', 'categories.id', '=', 'products.category_id')
                ->where('product_prices.level', 'Lista 1')
                ->whereNotNull('featured')
                // ->when(is_numeric($location_id), function($q) use($location_id){
                //     $q->where('product_stocks.location_id', '!=', $location_id);
                // })
                ->when(isset($categoryId), function($where) use($categoryId){
                    $where->where('category_id', $categoryId);
                })
                ->when(isset($brand), function($where) use($brand){
                    $where->where('brand', $brand);
                })
                ->validForEcommerce()
                ->groupBy('products.id')
                ->groupBy('products.name')
                ->groupBy('products.length')
                ->groupBy('products.width')
                ->groupBy('products.height')
                ->groupBy('products.weight')
                ->groupBy('products.featured')
                ->groupBy('products.data_sheet')
                ->groupBy('product_prices.price');

        if(is_numeric($location_id))
            $query->addSelect(DB::raw('
                (SELECT concat(locations.id, "||", ps.quantity_available , "||", locations.netsuite_name)
                FROM product_stocks AS ps
                JOIN locations on locations.id = ps.location_id
                WHERE ps.location_id = ' . $location_id . '
                AND ps.product_id = products.id
                AND ps.quantity_available > 0
                LIMIT 1) AS localStock'));

        return $query;
    }
    /**
     * Dar formato a una coleccion de productos paginados para ser
     * mostrados en la tienda
     *
     */
    public static function formatForIndex(\Illuminate\Pagination\LengthAwarePaginator $products){

        $products->getCollection()->transform(function ($product) {
            $product->featured_url = Storage::url($product->featured);
            $product->data_sheet_url = $product->data_sheet ? Storage::url($product->data_sheet) : null;

            // Aquí agregamos la lógica del stock
            if($product->localStock){
                list($id, $quantity_available, $name) = explode('||', $product->localStock);
                $product->localStock = [
                    'id' => $id,
                    'name' => $name,
                    'quantity' => $quantity_available,
                ];
            }
            if($product->nationalStock)
                $product->nationalStock = ['id' => 0, 'name' => 'Nacional', 'quantity' => $product->nationalStock];

            return $product;
        });

        return $products;
    }
}
