<?php

namespace App\Models;

use App\Services\Contracts\Tracking;
use Illuminate\Database\Eloquent\Factories\HasFactory;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Support\Facades\Storage;
use App\Services\ShippingServiceFactory;
use Illuminate\Support\Facades\DB;
use App\UseCases\Shipment\Cancel;
use App\UseCases\Shipment\Tracking as ShipmentTracking;

class Shipment extends Model
{
    public const DISK_LABEL_PDF = 'shipment_label';

    use HasFactory;

    protected $guarded = [];
    /**
     * Ver toda la informacion de un envio
     * - Información del envio
     * - Embalaje
     * @return object
     */
    public static function fullQuery(){
        return self::select(
                    'shipments.id',
                    'shipments.order_id',
                    'shipments.customer_id',
                    'shipments.carrier',
                    'shipments.shipping_cost',
                    'shipments.origin_postal_code',
                    'shipments.origin_address',
                    'shipments.destination_postal_code',
                    'shipments.destination_address',
                    'shipments.is_local_shipping',
                    'shipments.courier_code',
                    'shipments.master_tracking_number',
                    'shipments.delivery_status',
                    'shipments.delivery_event',
                    'shipments.estimated_delivery_date',
                    'shipments.service_code',
                    'shipments.customer_address_id',
                    'shipments.location_id',
                    'shipments.origin_neighborhood',
                    'shipments.origin_state',
                    'shipments.origin_contact_phone',
                    'shipments.origin_city',
                    'shipments.origin_country',
                    'shipments.destination_neighborhood',
                    'destination_state',
                    'shipments.destination_contact_phone',
                    'shipments.destination_city',
                    'shipments.destination_country',
                    'shipments.origin_number_ext',
                    'shipments.origin_number_int',
                    'shipments.destination_number_ext',
                    'shipments.destination_number_int',
                    'shipments.created_at',
                    'shipments.updated_at',
                    'shipments.delivery_amount',
                    DB::raw('shipment_carrier_log.reason AS shipment_carrier_log_reason'),
                    DB::raw('shipment_carrier_log.carrier as shipment_carrier_log_carrier'),
                    DB::raw('shipment_carrier_log.service_code as shipment_carrier_log_service_code'),
                    DB::raw('shipment_carrier_log.shipping_cost as shipment_carrier_log_shipping_cost'))
                ->leftJoin('shipment_carrier_log', 'shipment_carrier_log.shipment_id', '=', 'shipments.id');
    }
    /**
     * Guadar el PDF guia proporcionado por la paqueteria
     *
     * @param int $shipmentId
     * @param string $data
     * @return void
     */
    public static function saveShipmentLabel($shipmentId, $data){

        $fileName = $shipmentId . '.pdf';

        return Storage::disk(self::DISK_LABEL_PDF)->put($fileName, $data);

    }

    public function getFullOriginNumber(){
        return implode(', ', [$this->origin_number_ext, $this->origin_number_int]);
    }
    public function getFullDestinationNumber(){
        return implode(', ', [$this->destination_number_ext, $this->destination_number_int]);
    }
    /**
     *Esta consulta devuelve todos los datos necesarios para poder enviar mediante una paqueteria
     *
     * @param integer $shipmentId
     * @return Shipment
     */
    public static function forShipping(int $shipmentId): Shipment{
        return self::select(
                    'shipments.id',
                    'shipments.customer_id',
                    'shipments.location_id',
                    'shipments.order_id',
                    'shipments.service_code',
                    'shipments.origin_postal_code',
                    'shipments.origin_address',
                    'shipments.origin_country',
                    'shipments.origin_city',
                    'shipments.origin_contact_phone',
                    'shipments.origin_state',
                    'shipments.origin_neighborhood',
                    'shipments.origin_number_ext',
                    'shipments.origin_number_int',
                    'shipments.destination_postal_code',
                    'shipments.destination_address',
                    'shipments.destination_country',
                    'shipments.destination_city',
                    'shipments.destination_contact_phone',
                    'shipments.destination_state',
                    'shipments.destination_neighborhood',
                    'shipments.destination_number_ext',
                    'shipments.destination_number_int',
                    DB::raw('customers.company_name as destination_contact_name'),
                    DB::raw('locations.contact_name as origin_contact_name'),
                    DB::raw('SUM(order_items.quantity) as total_items')
                )
                ->join('customers', 'customers.id', '=', 'shipments.customer_id')
                ->join('locations', 'locations.id', '=', 'shipments.location_id')
                ->join('order_items', 'order_items.order_id', '=', 'shipments.order_id')
                ->groupBy(
                    'shipments.id',
                    'shipments.customer_id',
                    'shipments.location_id',
                    'shipments.order_id',
                    'shipments.service_code',
                    'shipments.origin_postal_code',
                    'shipments.origin_address',
                    'shipments.origin_country',
                    'shipments.origin_city',
                    'shipments.origin_contact_phone',
                    'shipments.origin_state',
                    'shipments.origin_neighborhood',
                    'shipments.destination_postal_code',
                    'shipments.destination_address',
                    'shipments.destination_country',
                    'shipments.destination_city',
                    'shipments.destination_contact_phone',
                    'shipments.destination_state',
                    'shipments.destination_neighborhood',
                    'customers.company_name',
                    'locations.contact_name')
                ->findOrFail($shipmentId);
    }
    /**
     * Rastrear el estado de un envio segun la paqueteria
     *
     * @return object
     */
    public function tracking(){

        $factory = new ShippingServiceFactory;
        $shippingService = $factory->make($this->carrier);
        $handler = new ShipmentTracking($this, $shippingService);

        return $handler->handle();
    }
    /**
     * Guardar el resultado del rastreo del envio
     *
     * @param Tracking $result
     * @return object
     */
    public function updateDeliveryStatusByTracking(Tracking $result)
    {
        $this->delivery_status = $result->getCode();
        $this->delivery_event = $result->getDescription();
        $this->estimated_delivery_date = $result->getPromiseDeliveryDateTime();
        $this->save();
    }
    public function canCancel(){
        $factory = new ShippingServiceFactory;
        $shippingService = $factory->make($this->carrier);

        return $this->delivery_status ?
                $shippingService->canCancel($this->delivery_status)
                : false;
    }
    public function isCanceled(){
        $factory = new ShippingServiceFactory;
        $shippingService = $factory->make($this->carrier);

        return $this->delivery_status ?
                $shippingService->isCanceled($this->delivery_status)
                : false;
    }
    /**
     * Cancelar el envio con la paqueteria
     *
     * @param integer $userId
     * @return bool
     */
    public function cancel(int $userId){
        $factory = new ShippingServiceFactory;
        $hadler = new Cancel($userId, $this, $factory->make($this->carrier));

        return $hadler->handle();
    }
    public function fullAddressOrigin(){
        return "$this->origin_address $this->origin_number_ext $this->origin_number_int,
                $this->origin_neighborhood, $this->origin_postal_code, $this->origin_city, $this->origin_state";
    }
}
