<?php
namespace App\Services\Estafeta;

use App\Models\OrderItem;
use App\Models\Shipment;
use App\Services\Estafeta\Exceptions\Estafeta;
use App\Services\Estafeta\Objects\Auth\TokenCacheConfig;
use App\Services\Estafeta\Objects\Quote\Dimensions;
use App\Services\Estafeta\Objects\Quote\Package;
use App\Services\Estafeta\Objects\Quote\PackageInformation;
use App\Services\Estafeta\Objects\Quote\PackagingType;
use App\Services\Estafeta\Objects\Shipment\Address;
use App\Services\Estafeta\Objects\Shipment\Contact;
use App\Services\Estafeta\Objects\Shipment\Destination;
use App\Services\Estafeta\Objects\Shipment\Identification;
use App\Services\Estafeta\Objects\Shipment\ItemDescription;
use App\Services\Estafeta\Objects\Shipment\LabelDefinition;
use App\Services\Estafeta\Objects\Shipment\Location;
use App\Services\Estafeta\Objects\Shipment\LocationLabel;
use App\Services\Estafeta\Objects\Shipment\Merchandise;
use App\Services\Estafeta\Objects\Shipment\Merchandises;
use App\Services\Estafeta\Objects\Shipment\Pallet;
use App\Services\Estafeta\Objects\Shipment\ServiceConfiguration;
use App\Services\Estafeta\Objects\Shipment\SystemInformation;
use App\Services\Estafeta\Objects\Shipment\WaybillDocument;
use App\Services\Estafeta\Objects\Tracking\SearchType;
use App\Services\Estafeta\Request\Auth;
use App\Services\Estafeta\Request\QuotationMultiple;
use App\Services\Estafeta\Request\Quote;
use App\Services\Estafeta\Request\Shipment as RequestShipment;
use App\Services\Estafeta\Request\ShipmentMultiple as RequestShipmentMultiple;
use App\Services\Estafeta\Request\ShipQueryParams;
use App\Services\Estafeta\Request\Tracking;
use App\Services\Shipments\Contracts\Packaging;
use App\Services\Shipments\Packer;
use App\Services\Shipments\PerSend\ListPackaging;
use App\Services\Shipments\ShippingQuote;
use GuzzleHttp\Client;
use App\Services\Estafeta\Objects\Enum\TypeService;
use App\Services\Shipments\Contracts\Product;

class Factory{
    public static function token(Auth $auth, bool $refresh = False){
        $handler = new Token(
            $auth,
            new Client(),
            new TokenCacheConfig(TokenCacheConfig::TOKEN_DELIVERY),
            $refresh
        );

        return $handler->get();
    }

    public static function quotation(ShippingQuote $shipment, string $Origin): Quote{
        $listPackaging = $shipment->getListPackaging();

        if($listPackaging->total() > 1) throw new Estafeta("EL envio debe ser mediante multiples paquetes", 1);

        $allPackages = $listPackaging->packages();
        $singlePackage = $allPackages[0];
        /**
         * La paqueteria requiere que los datos sean en Kg y cm
         */
        $weight = $singlePackage->getWeight(1000);
        $length = $singlePackage->getLength(10);
        $width = $singlePackage->getWidth(10);
        $depth = $singlePackage->getDepth(10);

        return new Quote(
            $Origin,
            [$shipment->destination->cp],
            PackagingType::equivalent($weight, $length, $width, $depth),
            new Dimensions($weight, $length, $width, $depth)
        );

    }
    public static function quotationMultiple(ShippingQuote $shipment, string $Origin): QuotationMultiple{
        $packageInformation = self::quotationPackageInformation($shipment->getListPackaging());

        return new QuotationMultiple(
            $Origin,
            [$shipment->destination->cp],
            $packageInformation
        );
    }
    private static function quotationPackageInformation(ListPackaging $listPackaging): PackageInformation{
        $items = $listPackaging->packages();

        if($listPackaging->total() == 1) throw new Estafeta("EL envio debe ser para un solo paquete", 1);

        $packageInformation =  new PackageInformation;
        $callback = function(Packaging $item, int $counter){
            $weight = $item->getWeight(1000);
            $length = $item->getLength(10);
            $width = $item->getWidth(10);
            $depth = $item->getDepth(10);

            return new Package(
                $counter,
                new Dimensions(
                    $weight,
                    $length,
                    $width,
                    $depth
                )
            );
        };

        foreach ($items as $key => $item) {
            $packageInformation->add($callback($item, $key + 1));
        }
        return $packageInformation;
    }

    /**
     *
     *
     * @param string $trackingId
     * @return Tracking
     */
    public static function tracking(string $trackingId): Tracking{
        return new Tracking(
            SearchType::BY_GUIDE,
            [$trackingId]
        );
    }
    /**
     * Obtener el empaquetado de un envio
     *
     * @param integer $shipment_id
     * @return ListPackaging
     */
    public static function package(int $shipment_id): ListPackaging{
        return Packer::byShipment($shipment_id);
    }
    private static function merchandise(Packaging $packaging){
        return array_map(function(Product $orderItem){
            return new Merchandise(
                $orderItem->claveProductoServicioCP(),
                $orderItem->quantity(),
                $orderItem->claveUnidadPeso(),
                false
            );
        },$packaging->getItems());
    }
    /**
     * Envio para un paquete
     *
     * @param Shipment $shipment
     * @param Packaging $packaging
     * @return RequestShipment
     */
    public static function shipmentPackage(Shipment $shipment, Packaging $packaging): RequestShipment{

        return new RequestShipment(
            new ShipQueryParams('FILE_PDF', 'REQUEST', 'SYNC_INLINE'),
            new Identification(
                config('estafeta.suscriber_id'),
                config('estafeta.customer_number')
            ),
            new SystemInformation('AP01', config('app.name'), '1.1'),
            self::labelDefinitionForPackaging($shipment, $packaging, 1, $shipment->service_code)
        );
    }
    /**
     * Envio para multiples paquetes
     *
     * @param Shipment $shipment
     * @param array $allPackaging
     * @return RequestShipmentMultiple
     */
    public static function shipmentPackages(Shipment $shipment, array $allPackaging): RequestShipmentMultiple{
        $packagesPerSend = [];

        foreach ($allPackaging as $referenceNumber => $packaging) {
            $packagesPerSend[] = self::labelDefinitionForPackaging(
                                    $shipment,
                                    $packaging,
                                    $referenceNumber,
                                    TypeService::MULTIPLE_TERRESTIAL->value
                                );
        }

        return new RequestShipmentMultiple(
            new ShipQueryParams('FILE_PDF', 'REQUEST', 'SYNC_INLINE'),
            new Identification(
                config('estafeta.suscriber_id'),
                config('estafeta.customer_number')
            ),
            new SystemInformation('AP01', config('app.name'), '1.1'),
            $packagesPerSend
        );
    }
    private static function labelDefinitionForPackaging(
        Shipment $shipment,
        Packaging $packaging,
        int $referenceNumber,
        string $service_code): LabelDefinition{
        /**
         * La paqueteria requiere que los datos sean en Kg y cm
         */
        $weight = $packaging->getWeight(1000);
        $length = $packaging->getLength(10);
        $width = $packaging->getWidth(10);
        $height = $packaging->getDepth(10);

        return new LabelDefinition(
            new WaybillDocument(
                __('Total de productos: ') . count($packaging->getItems()),
                $referenceNumber,
                Pallet::fromDimensions($weight, $length, $width, $height)
            ),
            new ItemDescription(
                $weight,
                $height,
                $length,
                $width,
                new Merchandises($weight, 'KGM', self::merchandise($packaging))
            ),
            new ServiceConfiguration(
                1,
                $service_code,
                config('estafeta.sales_organization'),
                $shipment->origin_postal_code,
                false
            ),
            new LocationLabel(
                new Location(
                    new Contact(config('app.razon_social'), $shipment->origin_contact_name, $shipment->origin_contact_phone),
                    new Address(
                        $shipment->origin_number_ext,
                        $shipment->origin_address,
                        'C',
                        $shipment->origin_neighborhood,
                        $shipment->origin_postal_code
                    )
                ),
                new Destination(
                    new Location(
                        new Contact($shipment->destination_contact_name, $shipment->destination_contact_name, $shipment->destination_contact_phone),
                        new Address(
                            $shipment->destination_number_ext,
                            $shipment->destination_address,
                            'C',
                            $shipment->destination_neighborhood,
                            $shipment->destination_postal_code
                        )
                    )
                )
            )
        );
    }
}
