<?php

namespace App\Services\Estafeta;

use App\Services\Estafeta\Exceptions\Estafeta;
use App\Services\Estafeta\Objects\Quote\Config;
use App\Services\Estafeta\Response\Quote as ResponseQuote;
use App\Services\Estafeta\Response\QuotationMultiple as ResponseQuotationMultiple;
use App\Services\Shipments\ShippingQuote;
use App\UseCases\Api\HttpRequestExceptionHandler;
use GuzzleHttp\Client;
use Illuminate\Support\Facades\Log;
use App\Services\Estafeta\Request\QuotationMultiple as RequestQuotationMultiple;
use App\Services\Estafeta\Request\Quote as RequestQuote;
use GuzzleHttp\Exception\ClientException;

class Quote extends HttpRequestExceptionHandler{

    private const MAX_INTENTS = 2;
    private  int $intents = 0;

    public function __construct(
        private readonly Client $client,
        private readonly Config $config)
    {

    }
    /**
     *
     *
     * @param string $trackingId
     * @param string $token
     * @return array
     */
    public function handle(ShippingQuote $shipment): array{

        $quotes = [];
        $listPackaging = $shipment->getListPackaging();

        if($listPackaging->total() > 1) throw new Estafeta("Aun no tenemos configurado los envios multiples", 1);

        foreach ($shipment->addresses as $address) {

            $quotation = $listPackaging->total() > 1 ? Factory::quotationMultiple($shipment, $address->cp) : Factory::quotation($shipment, $address->cp);
            $token = Factory::token($this->config->auth);

            try {
                \Log::info(json_encode($quotation->array()));
                $response = $this->sendRequest($quotation, $token);

            } catch (\Throwable $th) {
                continue;
            }

            $bodyResult = json_decode($response->getBody());
            \Log::info(json_encode($bodyResult, JSON_PRETTY_PRINT));
            if($quotation instanceof RequestQuote){
                if(isset($bodyResult->Quotation)){
                    //Algunas veces Estafeta no devuelve ninguna cotizacion, sin embargo no es un error, aqui se controla
                    if(isset($bodyResult->Quotation[0]->Service))
                        $quotes[] = new ResponseQuote($bodyResult, $address, $shipment->destination);
                }
                else{
                    Log::error("Estafeta: " . json_encode($bodyResult));
                }
            }else{
                if(isset($bodyResult->MultipleQoutations)){
                    //Algunas veces Estafeta no devuelve ninguna cotizacion, sin embargo no es un error, aqui se controla
                    if(count($bodyResult->MultipleQoutations[0]->LoadingQuotations) > 0)
                        $quotes[] = new ResponseQuotationMultiple($bodyResult, $address, $shipment->destination);
                }
                else{
                    Log::error("Estafeta: " . json_encode($bodyResult));
                }
            }

        }

        if(count($quotes) == 0) throw new \Exception("Estafeta: Sin cotizaciones disponibles", 1);

        $quotation = count($quotes) == 1 ? $quotes[0] : $this->getCheapest($quotes);

        return $quotation->array();

    }
    private function sendRequest($quotation, string $token){

        try {

            $response = $this->client->post($this->getUrlByTotalPackages($quotation), [
                'headers' => [
                    'Authorization' => 'Bearer ' . $token,
                    'apiKey' => $this->config->auth->client_id
                ],
                'json' => $quotation->array()
            ]);

            if($response->getStatusCode() === 200) return $response;

            throw new Estafeta('La respuesta de la paqueteria es inconsistente ' . json_encode($response), 1);

        } catch (ClientException $e) {
            if ($e->getCode() === 401) {
                if($this->intents < self::MAX_INTENTS)
                    return $this->reintentAuthAndRequest($quotation);
            }

            $this->handleHttpRequestException($e, 'Quote', Estafeta::class, $quotation->array());
        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Quote', Estafeta::class, $quotation->array());
        }
    }
    private function getCheapest($quotations): ResponseQuotationMultiple|ResponseQuote{

        $cheapestQuotation = array_pop($quotations);

        $cheapestQuotation = array_reduce($quotations,
            function(ResponseQuotationMultiple|ResponseQuote $cheapestQuotation, ResponseQuotationMultiple|ResponseQuote $quotation){

                if($quotation->cost() < $cheapestQuotation->cost()) $cheapestQuotation = $quotation;

                return $cheapestQuotation;
            }, $cheapestQuotation);

        return $cheapestQuotation;
    }
    private function getUrlByTotalPackages(RequestQuotationMultiple|RequestQuote $quotation): string{
        if($quotation instanceof RequestQuote)
            return config('estafeta.cotizador.url.single');
        else
            return config('estafeta.cotizador.url.multiple');
    }
    private function reintentAuthAndRequest($quotation){
        $this->intents++;
        Log::info("Estafeta: Reintentando autenticacion, intentos ($this->intents)");
        $this->sendRequest($quotation, Factory::token($this->config->auth, true));
    }
}
