<?php
namespace App\Services\Estafeta;

use App\Exceptions\Handler;
use App\Models\Shipment as ModelsShipment;
use App\Services\Contracts\Shipment as ContractsShipment;
use App\Services\Estafeta\Exceptions\Estafeta;
use App\Services\Estafeta\Response\Shipment as Response;
use App\Services\Estafeta\Response\ShipmentMultiple as ResponseShipmentMultiple;
use GuzzleHttp\Client;
use Illuminate\Support\Facades\Storage;
use App\Services\Estafeta\Helpers\Request;
use App\Services\Estafeta\Objects\Shipment\Config;
use App\Services\Estafeta\Request\Shipment as RequestShipment;
use App\Services\Estafeta\Request\ShipmentMultiple as RequestShipmentMultiple;
use App\Services\Shipments\PerSend\ListPackaging;
use App\UseCases\Api\HttpRequestExceptionHandler;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ServerException;
use Illuminate\Support\Facades\Log;

class Shipment extends HttpRequestExceptionHandler{

    use Request;
    private const MAX_INTENTS = 2;
    private  int $intents = 0;

    public function __construct(
        private readonly Client $client,
        private readonly Config $config)
    {

    }
    /**
     *
     *
     * @param integer $shipmentId
     * @param string $token
     * @return Response
     */
    public function handle(int $shipmentId): ContractsShipment{
        $listPackaging = Factory::package($shipmentId);
        $shipment = ModelsShipment::forShipping($shipmentId);
        $allPackaging = $listPackaging->packages();

        $request = $this->shipmentByTotalPackages($listPackaging->total(), $shipment, $allPackaging);
        $token = Factory::token($this->config->auth);
        $response = $this->sendRequest($request, $token);
        $bodyResult = $this->decodeResponse($response);

        return $this->processResponse(
            $bodyResult,
            $shipmentId,
            $listPackaging,
            $request instanceof RequestShipmentMultiple);
    }
    private function sendRequest(RequestShipment|RequestShipmentMultiple $shipment, string $token){

        try {

            $url = $this->getUrlByShipmentType($shipment);

            $response = $this->client->post($url, [
                'headers' => [
                    'Authorization' => 'Bearer ' . $token,
                    'apiKey' => $this->config->auth->client_id
                ],
                'query' => $shipment->queryParams(),
                'json' => $shipment->array()
            ]);

            if($response->getStatusCode() == 201) return $response;

            throw new Estafeta('No es posible conectarse a la paqueteria', 1);
        } catch (ServerException $e) {
            if ($e->getCode() === 500) {
                $response = $e->getResponse();
                $result = $response ? json_decode($response->getBody()) : null;

                if($result && $result->error == 'Invalid Token' && $this->intents < self::MAX_INTENTS)
                    return $this->reintentAuthAndRequest($shipment);
            }

            $this->handleHttpRequestException(
                $e,
                'Shipment',
                Estafeta::class,
                [
                    'query' => $shipment->queryParams(),
                    'json' => $shipment->array()
                ]);
        } catch (\Throwable $e) {
            $this->handleHttpRequestException(
                $e,
                'Shipment',
                Estafeta::class,
                [
                    'query' => $shipment->queryParams(),
                    'json' => $shipment->array()
                ]);
        }
    }
    private function decodeResponse(\Psr\Http\Message\ResponseInterface $response): object {

        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Estafeta('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    private function getUrlByShipmentType(RequestShipment|RequestShipmentMultiple $shipment): string{

        if($shipment instanceof RequestShipment)
            return $shipment->isLTL() ? config('estafeta.label_rest.url.ltl.single') : config('estafeta.label_rest.url.package.single');
        else
            return config('estafeta.label_rest.url.package.multiple');
    }
    private function processResponse(
        object $bodyResult,
        int $shipmentId,
        ListPackaging $ListPackaging,
        bool $isMultiple): ContractsShipment{

        if ($isMultiple) {
            if(count($bodyResult->labelPetitionResults) > 0) {

                $response = new ResponseShipmentMultiple($bodyResult, $ListPackaging, $shipmentId);

                $this->saveShipmentLabel($shipmentId, $response->getPdfEncoded());

                return $response;
            }
        }else{
            if(count($bodyResult->labelPetitionResult->elements) > 0) {

                $response = new Response($bodyResult, $ListPackaging, $shipmentId);

                $this->saveShipmentLabel($shipmentId, $response->getPdfEncoded());

                return $response;
            }
        }

        throw new \Exception("Hubo un error al procesar la solicitud", 1);
    }
    private function saveShipmentLabel(int $shipmentId, $encodePdf): bool{
        try {

            $fileName = $shipmentId . '.pdf';

            return Storage::disk(ModelsShipment::DISK_LABEL_PDF)->put($fileName, base64_decode($encodePdf));
        } catch (\Throwable $th) {
            Handler::log($th);
        }

        return false;
    }
    private function shipmentByTotalPackages(
        int $total,
        ModelsShipment $shipment,
        array $allPackaging): RequestShipment|RequestShipmentMultiple{

        return $total > 1 ? Factory::shipmentPackages($shipment, $allPackaging)
                            : Factory::shipmentPackage($shipment, $allPackaging[0]);
    }
    private function reintentAuthAndRequest(RequestShipment|RequestShipmentMultiple $shipment){
        $this->intents++;
        Log::info("Estafeta: Reintentando autenticacion, intentos ($this->intents)");
        $this->sendRequest($shipment, Factory::token($this->config->auth, true));
    }
}
