<?php
namespace App\Services\Estafeta;

use App\Services\Estafeta\Exceptions\Estafeta;
use App\Services\Estafeta\Request\Auth;
use GuzzleHttp\Client;
use Illuminate\Support\Facades\Cache;
use App\Services\Estafeta\Objects\Auth\TokenCacheConfig;
use App\UseCases\Api\HttpRequestExceptionHandler;
use Illuminate\Support\Facades\Log;

/**
 * Obtener un token para poder realizar solicitudes a la Api
 * https://disenoweb.estafeta.com/WSCotizador/ES/2023_Manual%20WSCotizadorPrecioLista_Descuento_ES_V0.4_vDise%C3%B1oNvo.pdf
 */
class Token extends HttpRequestExceptionHandler{

    /**
     * Margen de seguridad para el tiempo de vida del token
     */
    const TOKEN_EXPIRATION_MARGIN = 10;
    /**
     * Máximo número de intentos de reintento en caso de fallo
     */
    const MAX_RETRIES = 2;

    public function __construct(
        private Auth $auth,
        private readonly Client $client,
        private readonly TokenCacheConfig $tokenCacheConfig,
        private readonly bool $refresh = false)
    {

    }
    public function get(): string {

        if($this->refresh) $this->getInApi();

        $token = $this->getInCache();
        return $token ? $token : $this->getInApi();

    }
    private function getInApi(): string {
        $attempts = 0;
        // Lógica de reintento
        while ($attempts < self::MAX_RETRIES) {
            try {

                $response = $this->sendRequest();
                $bodyResult = $this->decodeResponse($response);
                return $this->processResponse($bodyResult);

            } catch (\Throwable $e) {

                $attempts++;

                if ($attempts >= self::MAX_RETRIES) {
                    throw new \Exception("Máximo de reintentos de la solicitud", 1);
                } else {
                    Log::warning("Reintentando solicitud de token a la API de Estafeta. Intento: $attempts");

                    sleep(pow(2, $attempts)); // Retraso exponencial antes de reintentar
                }
            }
        }
    }

    private function sendRequest(): \Psr\Http\Message\ResponseInterface{
        try {

            $response = $this->client->post($this->auth->url,
                    [
                        'headers' => ['content-type' => 'application/x-www-form-urlencoded'],
                        'form_params' => [
                            'grant_type' => $this->auth->grant_type,
                            'client_id' => $this->auth->client_id,
                            'client_secret' => $this->auth->client_secret,
                            'scope' => $this->auth->scope
                        ]
                    ]
                );

            if($response->getStatusCode() === 200) return $response;

            throw new Estafeta('No es posible conectarse a la paqueteria: ' . json_encode($response), 1);

        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Token', Estafeta::class);
        }
    }
    private function decodeResponse(\Psr\Http\Message\ResponseInterface $response): object{
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Estafeta('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    private function processResponse(object $bodyResult): string{

        if (!isset($bodyResult->access_token) || !isset($bodyResult->expires_in))
            throw new Estafeta('La respuesta de la API no contiene el token o tiempo de expiración: ' . json_encode($bodyResult));

        $this->saveInCache($bodyResult->access_token, $bodyResult->expires_in);

        return $bodyResult->access_token;
    }
    private function getInCache():?string {
        return  Cache::get($this->tokenCacheConfig->token);
    }
    private function saveInCache(string $token, int $expires_in): void {
        Cache::set($this->tokenCacheConfig->token, $token, now()->addSeconds($this->getTokenSafeExpirationTime($expires_in)));
    }
    public function getTokenSafeExpirationTime(int $expires_in): int {
        return $expires_in - self::TOKEN_EXPIRATION_MARGIN;
    }
}
