<?php
namespace App\Services\Fedex;

use App\Models\Shipment as ModelsShipment;
use App\Services\Fedex\Helpers\Formatters;
use App\Services\Fedex\Objects\Dimensions;
use App\Services\Fedex\Request\Auth;
use App\Services\Fedex\Request\Credencials;
use GuzzleHttp\Client;
use App\Services\Fedex\Objects\Quote\Address;
use App\Services\Fedex\Objects\PackageLineItem;
use App\Services\Fedex\Objects\PickupType;
use App\Services\Fedex\Objects\RequestedPackageLineItems;
use App\Services\Fedex\Objects\Quote\RequestedShipment;
use App\Services\Fedex\Objects\ServiceType;
use App\Services\Fedex\Objects\Shipment\Address as ShipmentAddress;
use App\Services\Fedex\Objects\Shipment\Contact;
use App\Services\Fedex\Objects\Shipment\LabelSpecification;
use App\Services\Fedex\Objects\Shipment\PaymentType;
use App\Services\Fedex\Objects\Shipment\Recipient;
use App\Services\Fedex\Objects\Shipment\RequestedShipment as ShipmentRequestedShipment;
use App\Services\Fedex\Objects\Shipment\Shipper;
use App\Services\Fedex\Objects\Shipment\ShippingChargesPayment;
use App\Services\Fedex\Objects\SubPackagingType as ObjectsSubPackagingType;
use App\Services\Fedex\Objects\TokenCacheConfig;
use App\Services\Fedex\Objects\Track\AssociatedType;
use App\Services\Fedex\Objects\Track\MasterTrackingNumberInfo;
use App\Services\Fedex\Objects\Track\TrackingNumberInfo;
use App\Services\Fedex\Request\Cancel;
use App\Services\Fedex\Request\Rate;
use App\Services\Fedex\Request\Shipment;
use App\Services\Fedex\Request\Track;
use App\Services\Shipments\ShippingQuote;
use App\Services\Shipments\Contracts\Packaging;
use App\Services\Shipments\PerSend\ListPackaging;
use App\Services\Shipments\Contracts\Product as ContractProduct;

class Factory{

    public static function token(bool $refresh = false): string{
        $token = new Token(
            new Auth(
                config('fedex.key'),
                config('fedex.secret_key')
            ),
            new Client(['base_uri' => config('fedex.base_uri')]),
            new TokenCacheConfig(TokenCacheConfig::TOKEN_DELIVERY),
            $refresh
        );

        return $token->get();
    }
    public static function credentials(bool $refresh = false): Credencials{
        return new Credencials(
            self::token($refresh)
        );
    }
    public static function tokenTracking(bool $refresh = false): string{

        $token = new Token(
            new Auth(
                config('fedex.key_tracking'),
                config('fedex.secret_key_tracking')
            ),
            new Client(['base_uri' => config('fedex.base_uri')]),
            new TokenCacheConfig(TokenCacheConfig::TOKEN_TRACKING),
            $refresh
        );

        return $token->get();
    }
    public static function credentialsTracking(bool $refresh = false): Credencials{
        return new Credencials(
            self::tokenTracking($refresh)
        );
    }
    public static function rate(string $origin, string $destination, RequestedPackageLineItems $requestedPackageLineItems): Rate{

        return new Rate(
            config('fedex.account'),
            new RequestedShipment(
                new Address(
                    $origin
                ),
                new Address(
                    $destination
                ),
                new PickupType(PickupType::CONTACT_FEDEX_TO_SCHEDULE),
                $requestedPackageLineItems
            )
        );
    }

    public static function service():Service{
        return new Service(
            new Client(['base_uri' => config('fedex.base_uri')])
        );
    }
    /**
     *
     *
     * @param integer $shipmentId
     * @return Shipment
     */
    public static function shipment(int $shipmentId, RequestedPackageLineItems $requestedPackageLineItems): Shipment{

        $shipment = ModelsShipment::forShipping($shipmentId);

        return new Shipment(
                new ShipmentRequestedShipment(
                    new Shipper(
                        new ShipmentAddress(
                            Formatters::streetLines($shipment->origin_address, $shipment->origin_number_ext, $shipment->origin_number_int),
                            $shipment->origin_city,
                            $shipment->origin_postal_code
                        ),
                        new Contact($shipment->origin_contact_phone, $shipment->origin_contact_name, $shipment->origin_contact_name)
                    ),
                    [new Recipient(
                        new ShipmentAddress(
                                Formatters::streetLines($shipment->destination_address, $shipment->destination_number_ext, $shipment->destination_number_int),
                                $shipment->destination_city,
                                $shipment->destination_postal_code
                        ),
                        new Contact($shipment->destination_contact_phone, $shipment->destination_contact_name, $shipment->destination_contact_name)
                    )],
                    new PickupType(PickupType::CONTACT_FEDEX_TO_SCHEDULE),
                    new ServiceType($shipment->service_code),
                    new ShippingChargesPayment(new PaymentType(PaymentType::SENDER)),
                    new LabelSpecification(),
                    $requestedPackageLineItems
                ),
                config('fedex.account')
            );
    }
    public static function track(string $trackingId): Track{
        return new Track(
            new MasterTrackingNumberInfo(
                new TrackingNumberInfo(
                    $trackingId
                )
                ),
            AssociatedType::STANDARD_MPS
        );
    }
    public static function cancel($trackingNumber){
        return new Cancel(config('fedex.account'), $trackingNumber);
    }
    /**
     *
     *
     * @param ShippingQuote $shipment
     * @return RequestedPackageLineItems
     */
    public static function requestedPackageLineItemsByShipment(ShippingQuote $shipment): RequestedPackageLineItems{
        $lineItems = new RequestedPackageLineItems;

        $items = $shipment->getListPackaging()->packages();
        foreach ($items as $key => $item) {
            $lineItems->add(self::createRequestedPackageLineItems($item, ($key + 1)));
        }

        return $lineItems;
    }
    /**
     *
     *
     * @param ShippingQuote $shipment
     * @return RequestedPackageLineItems
     */
    public static function requestedPackageLineItemsByPackaging(ListPackaging $listPackaging): RequestedPackageLineItems{
        $lineItems = new RequestedPackageLineItems;

        $items = $listPackaging->packages();
        foreach ($items as $key => $item) {
            $lineItems->add(self::createRequestedPackageLineItems($item, ($key + 1)));
        }

        return $lineItems;
    }
    private static function createRequestedPackageLineItems(Packaging $packaging, int $groupPackageCount): PackageLineItem {
        /**
        * La paqueteria requiere que los datos sean en Kg y cm
        */
        $weight = $packaging->getWeight(1000);
        $length = $packaging->getLength(10);
        $width = $packaging->getWidth(10);
        $depth = $packaging->getDepth(10);
        //Usamos el item cuyo subtotal sea el mayor y lo usamos para especificar el contenido del paquete
        $itemMostExpense = array_reduce($packaging->getItems(), function(ContractProduct|null $itemMostExpense, ContractProduct $orderItem){

            if($itemMostExpense === null || $orderItem->subtotal() > $itemMostExpense->subtotal())
                $itemMostExpense = $orderItem;

            return $itemMostExpense;
        });

        $subPackagingType = ObjectsSubPackagingType::equivalent($weight, $length, $width, $depth);
        // if ($subPackagingType->value == ObjectsSubPackagingType::PALLET)
        //     throw new \Exception('Fedex no tiene un contrato de precio especial con Solar Center para envios pesados', 1);

        return new PackageLineItem(
            $weight,
            $subPackagingType,
            new Dimensions($length, $width, $depth),
            $itemMostExpense->getDescription(),
            $groupPackageCount
        );
    }
}
