<?php
namespace App\Services\Fedex\Objects\Quote;

use App\Services\Fedex\Objects\PickupType;
use App\Services\Fedex\Objects\RequestedPackageLineItems;

/**
 * https://developer.fedex.com/api/es-mx/catalog/rate/v1/docs.html
 */
class RequestedShipment{

    /**
     * @param Address $shipper The descriptive data for the physical shipper location from which the shipment originates.
     * @param Address $recipient Indicate the descriptive data for the recipient location to which the shipment is to be received.
     * @param PickupType $pickupType Indicate the pickup type method by which the shipment to be tendered to FedEx.
     * @param RequestedPackageLineItems $requestedPackageLineItems These are one or more package-attribute descriptions, each of which describes an individual package,
     * a group of identical packages, or (for the total-piece-total-weight case) common characteristics of all packages in the shipment.
     * @param string $preferredCurrency Indicate the currency the caller requests to have used in all returned monetary values (when a choice is possible).
     * Used in conjunction with the rateRequestType data element. This element is used to pull Preferred rates.
     * https://developer.fedex.com/api/es-mx/guides/api-reference.html#currencycodes
     * @param string $rateRequestType Indicate the type of rates to be returned.
     *Following are values:
     *LIST - Returns FedEx published list rates in addition to account-specific rates (if applicable).
     *PREFERRED - Returns rates in the preferred currency specified in the element preferredCurrency.
     *ACCOUNT - Returns account specific rates (Default).
     *INCENTIVE - This is one-time discount for incentivising the customer. For more information, contact your FedEx representative.
     * @param string $packagingType https://developer.fedex.com/api/es-mx/guides/api-reference.html#packagetypes
     */
    public function __construct(
        public readonly Address $shipper,
        public readonly Address $recipient,
        public readonly PickupType $pickupType,
        public readonly RequestedPackageLineItems $requestedPackageLineItems,
        public readonly string $preferredCurrency = 'NMP',
        public readonly array $rateRequestType = ['ACCOUNT'],
        public readonly string $packagingType = 'YOUR_PACKAGING'
        ){}

    public function array(){
        return [
            'shipper' => [
                'address' => $this->shipper->array()
            ],
            'recipient' => [
                'address' => $this->recipient->array()
            ],
            'pickupType' => $this->pickupType->value,
            'requestedPackageLineItems' => $this->requestedPackageLineItems->array(),
            'preferredCurrency' => $this->preferredCurrency,
            'rateRequestType' => $this->rateRequestType,
            'packagingType' => $this->packagingType
        ];
    }

}
