<?php
namespace App\Services\Fedex;

use App\Services\Fedex\Exceptions\Fedex;
use App\Services\Fedex\Helpers\Request;
use GuzzleHttp\Client;
use App\Services\Fedex\Response\Rate;
use App\Services\Fedex\Request\Rate as RequestRate;
use App\Services\Shipments\ShippingQuote;
use App\UseCases\Api\HttpRequestExceptionHandler;
use GuzzleHttp\Exception\ClientException;
use Illuminate\Support\Facades\Log;

class Quote extends HttpRequestExceptionHandler{

    use Request;
    private const MAX_INTENTS = 2;
    private int $intents = 0;
    private const API_URL = '/rate/v1/rates/quotes';

    public function __construct(private readonly Client $client)
    {

    }
    public function handle(ShippingQuote $shipment){
        $packageLineItemsByShipment = Factory::requestedPackageLineItemsByShipment($shipment);
        $rates = [];

        foreach ($shipment->addresses as $address) {

            $request = Factory::rate($address->cp, $shipment->destination->cp, $packageLineItemsByShipment);

            try {

                $response = $this->sendRequest($request);

            } catch (\Throwable $th) {
                continue;
            }

            $bodyResult = $this->decodeResponse($response);

            $rates[] = new Rate($bodyResult, $address, $shipment->destination);

        }

        if(count($rates) === 0) throw new Fedex('Sin ninguna cotizacion de Fedex', 1);

        $rate = count($rates) == 1 ? $rates[0] : $this->selectBestQuote($rates);

        return $rate->array();
    }
    private function sendRequest(RequestRate $request){
        $headers = Factory::credentials($this->intents > 0);

        try {

            $response = $this->client->post(self::API_URL, [
                'headers' =>[
                    'Authorization' => 'Bearer ' . $headers->authorization,
                    'X-locale' => $headers->xLocale,
                    'Content-Type' => $headers->contentType
                ],
                'json' => $request->array()
            ]);

            if($response->getStatusCode() === 200) return $response;

            throw new Fedex('La respuesta de la paqueteria es inconsistente ' . json_encode($response), 1);

        }
        catch (ClientException $e) {
            if ($e->getCode() === 401) {
                if($this->intents < self::MAX_INTENTS)
                    return $this->reintentAuthAndRequest($request);
            }

            $this->handleHttpRequestException($e, 'Quote', Fedex::class, $request->array());
        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Quote', Fedex::class, $request->array());
        }
    }
    private function decodeResponse(\Psr\Http\Message\ResponseInterface $response): object{
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Fedex('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    private function selectBestQuote($rates){

        $cheapestQuote = array_pop($rates);

        $cheapestQuote = array_reduce($rates,  function($cheapestQuote, $rate){

            if($rate->cheapestRate->totalNetCharge < $cheapestQuote->cheapestRate->totalNetCharge) $cheapestQuote = $rate;

            return $cheapestQuote;
        }, $cheapestQuote);

        return $cheapestQuote;
    }
    private function reintentAuthAndRequest(RequestRate $request){
        $this->intents++;
        Log::info("Fedex: Reintentando autenticacion, intentos ($this->intents)");
        $this->sendRequest($request);
    }
}
