<?php
namespace App\Services\Fedex\Response;

use App\Services\Fedex\Objects\ServiceType;
use App\Services\Helpers\Money;
use App\Services\Shipments\Address;

class Rate{

    /**
     * Moneda de respaldo en caso de que la del usuario no este disponible,
     * segun los codigos de Fedex https://developer.fedex.com/api/es-mx/guides/api-reference.html#currencycodes
     */
    private const DEFAULT_CURRENCY = 'USD';
    public object $cheapestQuote;
    public object $shipmentDetail;
    public ?object $cheapestRate;
    /**
     *
     *
     * @param object $data Respuesta desde FEDEX  https://developer.fedex.com/api/es-mx/catalog/rate/v1/docs.html
     * @param string $currencyCustomer Moneda del usuario segun los codigos de Fedex https://developer.fedex.com/api/es-mx/guides/api-reference.html#currencycodes
     */
    public function __construct(
        public readonly object $data,
        public readonly Address $origin,
        public readonly Address $destination,
        private readonly string $currencyCustomer = 'NMP'){

        if(!$this->getCheapestQuote($currencyCustomer)){

            if(!$this->getCheapestQuote(self::DEFAULT_CURRENCY)) throw new \Exception('No se encontro una cotizacion en dólares o pesos mexicanos', 1);

        }
    }

    public function array(){
        return [
            'Quotation' => [
                [
                    'Service' => [
                        [
                            'ServiceCode' => $this->cheapestQuote?->serviceType,
                            'ServiceName' => $this->cheapestQuote?->serviceDescription->description,
                            'Modality' => $this->cheapestQuote?->serviceDescription?->serviceId,
                            'TotalAmount' => $this->cheapestRate->currency == self::DEFAULT_CURRENCY ?
                                                Money::convertToPeso($this->cheapestRate?->totalNetCharge) : $this->cheapestRate?->totalNetCharge
                        ],
                    ],
                    'Origin' => [
                        'Postalcode' => $this->origin->cp,
                        'WarehouseCode' => 'N/A',
                        'TownshipName' => 'N/A', // No proporcionado
                        'StateName' => 'N/A', // No proporcionado
                    ],
                    'Destination' => [
                        'Postalcode' => $this->destination->cp,
                        'WarehouseCode' => 'N/A',
                        'TownshipName' => 'N/A', // No proporcionado
                        'StateName' => 'N/A', // No proporcionado
                    ],
                    'Message' => $this->data ?? 'Ok',
                ],
            ],
            'origin' => $this->origin,
            'destination' => $this->destination,
            'carrier' => 'fedex'
        ];
    }
    /**
     * Hay que obtener la cotizacion con la tarifa más barata,
     * sin embargo fedex nos responde la misma cotizacino en diferentes monedas, hay que tener eso en cuenta
     *
     * @return bool
     */
    private function getCheapestQuote(string $preferredCurrency): bool
    {
        $quotes = array_filter($this->data->output->rateReplyDetails, fn($quote) => ServiceType::isValid($quote->serviceType));

        if (empty($quotes)) {
            return false;
        }

        $this->cheapestQuote = $quotes[0];
        $this->cheapestRate = $this->getLowestRate($this->cheapestQuote->ratedShipmentDetails, $preferredCurrency);

        foreach ($quotes as $quote) {
            $rate = $this->getLowestRate($quote->ratedShipmentDetails, $preferredCurrency);
            if ($rate && $rate->totalNetCharge < $this->cheapestRate->totalNetCharge) {
                $this->cheapestQuote = $quote;
                $this->cheapestRate = $rate;
            }
        }

        return $this->cheapestRate !== null;
    }
    /**
     * Fedex incluye multiples tarifas dentro de cada opcion segun el tipo de moneda, sigue siendo el mismo servicio
     * pero con diferente moneda
     *
     * @param array $rateDetails
     * @param string $currency
     * @return object|null
     */
    private function getLowestRate(array $rateDetails, string $currency): ?object
    {
        return array_reduce($rateDetails, function($lowest, $rateDetail) use ($currency) {
            if ($rateDetail->currency === $currency) {
                return !$lowest || $rateDetail->totalNetCharge < $lowest->totalNetCharge ? $rateDetail : $lowest;
            }
            return $lowest;
        }, null);
    }
}
