<?php
namespace App\Services\Fedex\Response;

use App\Services\Contracts\Tracking as ContractsTracking;
use App\Services\Fedex\Exceptions\Fedex;

/**
 * Ver https://developer.fedex.com/api/es-us/catalog/track/v1/docs.html
 */
class Tracking implements ContractsTracking{

    private $latestStatusDetail;
    private $trackResult;
    private $status;

    public function __construct(public readonly object $response)
    {
        if (!isset($this->response->output->completeTrackResults[0]->trackResults[0])) {
            throw new Fedex('La respuesta no contiene resultados de seguimiento válidos.');
        }
        $this->latestStatusDetail = $this->response->output->completeTrackResults[0]->trackResults[0]->latestStatusDetail;
        $this->trackResult = $this->response->output->completeTrackResults[0]->trackResults[0];
        $this->status = $this->trackResult->scanEvents[count($this->trackResult->scanEvents) - 1];
    }
    public function getCode(): string
    {
        return $this->status->eventType;
    }
    public function getDescription(): string
    {
        return __('fedex.status-tracking-' . $this->status->eventType);
    }
    public function getDelay(): ?string{
        $status = $this->latestStatusDetail?->delayDetail?->status ?? null;

        if($status == 'DELAYED')
            return __('fedex.status-tracking-delay-type-' . $this->latestStatusDetail->delayDetail->type);

        return null;
    }
    public function getDateTime(): ?string{
        $date = new \DateTime($this->status->date);

        return $date->format('d/m/Y H:i:s A');
    }
    public function getPromiseDeliveryDateTime(): ?string{

        $window = $this->trackResult->estimatedDeliveryTimeWindow->window ?? null;

        if ($window && isset($window->begins, $window->ends)) {
            $beginDate = (new \DateTime($window->begins))->format('d/m/Y H:i:s A');
            $endDate = (new \DateTime($window->ends))->format('d/m/Y H:i:s A');
            return __('fedex.status-promise-delivery-dateTime', ['begin' => $beginDate, 'end' => $endDate]);
        }
        return null;
    }
    public function getAdditionalInfo(): ?string{
        $rows = '';

        $rows .= $this->getFormattedInformationNotes();
        $rows .= $this->getFormattedAncillaryDetails();
        $rows .= $this->getFormattedAlerts();

        return strlen($rows) > 0 ?  "<table class='table table-sm'>
                    $rows
                </table>" : null;

    }
    public function getHistorical(): ?string{
        // Omitimos el último evento ya que es el estado actual.
        $length = count($this->trackResult->scanEvents) - 2;

        if ($length >= 0) {

            $groupedData = $this->historicalGroupByDate();
            $rows = '';

            foreach ($groupedData as $date => $events) {

                $rows .= view('components.tracking.tr_date_group', [ 'date' => $date])->render();
                $rows .= $this->rowsEvents($events);

            }

            return "<div>
                        <table class='table table-sm'>
                            $rows
                        </table>
                    </div>";

        }
        return null;
    }
    private function rowsEvents(array $events): string{
        $rows = '';
        foreach ($events as $event) {

            $date = new \DateTime($event->date);
            $hour = $date->format('h:i a');
            $description = __('fedex.status-tracking-' . $event->eventType);
            $location = isset($event->scanLocation) ? $event->scanLocation->city . implode(',', $event->scanLocation->streetLines) : 'Desconocido';

            $rows .= view('components.tracking.tr', ['hour' => $hour, 'description' => $description, 'location' => $location])
                        ->render();
        }
        return $rows;
    }
    private function historicalGroupByDate(): array{

        $groupedData = [];

        foreach ($this->trackResult->scanEvents as $event) {
            $date = new \DateTime($event->date);
            $formatDate = $date->format('d/m/Y');
            if (!isset($groupedData[$formatDate])) {
                $groupedData[$formatDate] = [];
            }
            $groupedData[$formatDate][] = $event;
        }

        return $groupedData;
    }
    private function getFormattedInformationNotes(): string
    {
        return $this->informationNotes($this->trackResult?->informationNotes ?? []);
    }

    private function getFormattedAncillaryDetails(): string
    {
        return $this->ancillaryDetails($this->latestStatusDetail?->ancillaryDetails ?? []);
    }

    private function getFormattedAlerts(): string
    {
        return $this->alerts($this->response->output->alerts ?? []);
    }
    private function informationNotes(array $informationNotes): string {
        return array_reduce($informationNotes, function($rows, $info){
                $rows .= "<tr>
                            <td>$info->description</td>
                        </tr>";
                return $rows;
            }, '');
    }
    private function ancillaryDetails(array $ancillaryDetails): string{
        return array_reduce($ancillaryDetails, function($rows, $details){

            $action = $details->action ? "<div class='bg-primary text-white'>Acción requerida: $details->action</div>" : '';

            $rows .= "<tr>
                        <td>
                            <div>$details->reasonDescription</div>
                            $action
                        </td>
                    </tr>";

            return $rows;

        }, '');
    }
    private function alerts(array $alerts): string{
        return array_reduce($alerts, function($rows, $alert){
                $rows .= "<tr>
                            <td>$alert->message</td>
                        </tr>";
                return $rows;
            }, '');
    }
}
