<?php

namespace App\Services\Fedex;

use App\Exceptions\Handler;
use App\Services\Fedex\Exceptions\Fedex;
use App\Services\Fedex\Helpers\Request;
use GuzzleHttp\Client;
use App\Services\Fedex\Request\Shipment as RequestShipment;
use App\Services\Fedex\Response\Shipment as ResponseShipment;
use App\Services\Shipments\Packer;
use App\Services\Shipments\PerSend\ListPackaging;
use App\UseCases\Api\HttpRequestExceptionHandler;
use Illuminate\Support\Facades\Log;
use GuzzleHttp\Exception\ClientException;

class Shipment extends HttpRequestExceptionHandler{

    use Request;
    private const MAX_INTENTS = 2;
    private int $intents = 0;
    private const API_URL = '/ship/v1/shipments';
    public function __construct(private readonly Client $client)
    {

    }
    /**
    * Solicitar un envio, fedex recomienda primero validar que las direcciones y estructura de la solicitd sea valida
    * asi que primero lo hacemos y luego solicitamos el envio(De momento no tenemos convenio para envios LTL)
    *
    * @param integer $shipmentId
    * @return ResponseShipment
    */
    public function handle(int $shipmentId): ResponseShipment{

        $listPackaging = Packer::byShipment($shipmentId);
        $requestedPackageLineItems = Factory::requestedPackageLineItemsByPackaging($listPackaging);
        $request = Factory::shipment($shipmentId, $requestedPackageLineItems);

        //$this->validateShipment($request); Debido a que la respuesta es vitualizada para pruebas lo elimine, ya que puede fallar de manera aleatoria

        $response = $this->sendRequest($request);
        $bodyResult = $this->decodeResponse($response);

        return $this->processResponse($bodyResult, $listPackaging, $shipmentId);
    }
    private function validateShipment(RequestShipment $request){

        $handler = new ValidateShipment($this->client);

        return $handler->handle($request);
    }
    private function sendRequest(RequestShipment $request){
        $headers = Factory::credentials($this->intents > 0);

        try {

            $response = $this->client->post(self::API_URL, [
                'headers' =>[
                    'Authorization' => 'Bearer ' . $headers->authorization,
                    'X-locale' => $headers->xLocale,
                    'Content-Type' => $headers->contentType
                ],
                'json' => $request->array()
            ]);

            if($response->getStatusCode() === 200)  return $response;

            throw new Fedex('La respuesta de la paqueteria es inconsistente ' . json_encode($response), 1);

        }
        catch (ClientException $e) {
            if ($e->getCode() === 401) {
                if($this->intents < self::MAX_INTENTS)
                    return $this->reintentAuthAndRequest($request);
            }

            $this->handleHttpRequestException($e, 'Shipment', Fedex::class, $request->array());
        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Shipment', Fedex::class, $request->array());
        }
    }
    private function decodeResponse(\Psr\Http\Message\ResponseInterface $response): object{
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Fedex('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    private function processResponse(object $bodyResult, ListPackaging $listPackaging, int $shipmentId): ResponseShipment{

        $response = new ResponseShipment($bodyResult, $listPackaging, $shipmentId);

        $this->downloadPdf($response->getUrlPdfForDownload(), $shipmentId);

        return $response;
    }
    private function downloadPdf(string $url, int $shipmentId){
        $handler = new ShipmentLabel($this->client);

        try {

            $handler->handle($url, $shipmentId);

        } catch (\Throwable $th) {
            Log::info("No se pudo descargar el pdf del envio $shipmentId, url: $url");
            Handler::log($th);
        }
    }
    private function reintentAuthAndRequest(RequestShipment $request){
        $this->intents++;
        Log::info("Fedex: Reintentando autenticacion, intentos ($this->intents)");
        $this->sendRequest($request);
    }
}
