<?php
namespace App\Services\Fedex;

use App\Services\Fedex\Exceptions\Fedex;
use App\Services\Fedex\Request\Auth;
use GuzzleHttp\Client;
use Illuminate\Support\Facades\Cache;
use App\Services\Fedex\Helpers\Request;
use App\Services\Fedex\Objects\TokenCacheConfig;
use App\UseCases\Api\HttpRequestExceptionHandler;
use Illuminate\Support\Facades\Log;

/**
 * Obtener un token para poder realizar solicitudes a la Api
 * https://developer.fedex.com/api/es-mx/catalog/authorization/v1/docs.html#operation/API%20Authorization
 */
class Token extends HttpRequestExceptionHandler{

    use Request;
    private const API_URL = '/oauth/token';
    /**
     * Margen de seguridad para el tiempo de vida del token
     */
    const TOKEN_EXPIRATION_MARGIN = 10;
    /**
     * Máximo número de intentos de reintento en caso de fallo
     */
    const MAX_RETRIES = 2;

    public function __construct(
        private Auth $auth,
        private readonly Client $client,
        private readonly TokenCacheConfig $tokenCacheConfig,
        private readonly bool $refresh = false)
    {

    }
    public function get(): string {

        if($this->refresh) $this->getInApi();

        $token = $this->getInCache();
        return $token ?  $token : $this->getInApi();

    }

    private function getInApi(): string {
        $attempts = 0;
        // Lógica de reintento
        while ($attempts < self::MAX_RETRIES) {
            try {

                $response = $this->sendRequest();
                $bodyResult = $this->decodeResponse($response);
                return $this->processResponse($bodyResult);

            } catch (\Throwable $e) {

                $attempts++;

                if ($attempts >= self::MAX_RETRIES) {
                    throw new \Exception("Máximos de reintentos alcanzados", 1);
                } else {
                    Log::warning("Reintentando solicitud de token a la API de Fedex. Intento: $attempts");

                    sleep(pow(2, $attempts)); // Retraso exponencial antes de reintentar
                }
            }
        }
    }
    private function sendRequest(): \Psr\Http\Message\ResponseInterface{

        try {

            $response = $this->client->post(self::API_URL,
            [
                'headers' => ['content-type' => $this->auth->contentType],
                'form_params' => [
                    'client_id' => $this->auth->client_id,
                    'client_secret' => $this->auth->client_secret,
                    'grant_type' => $this->auth->grant_type
                ]
            ]
        );

        if($response->getStatusCode() === 200) return $response;

        throw new Fedex('La respuesta de la paqueteria es inconsistente ' . json_encode($response), 1);

        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Shipment', Fedex::class);
        }
    }
    private function decodeResponse(\Psr\Http\Message\ResponseInterface $response): object{
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Fedex('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    private function processResponse(object $bodyResult): string{

        $this->saveInCache($bodyResult->access_token, $bodyResult->expires_in);

        return $bodyResult->access_token;
    }
    private function getInCache():?string {
        return  Cache::get($this->tokenCacheConfig->token);
    }
    private function saveInCache(string $token, int $expires_in): void {
        Cache::set($this->tokenCacheConfig->token, $token, now()->addSeconds($this->getTokenSafeExpirationTime($expires_in)));
    }
    public function getTokenSafeExpirationTime(int $expires_in): int {
        return $expires_in - self::TOKEN_EXPIRATION_MARGIN;
    }
}
