<?php

namespace App\Services\Fedex;

use App\Services\Fedex\Exceptions\Fedex;
use App\Services\Fedex\Helpers\Request;
use GuzzleHttp\Client;
use App\Services\Fedex\Request\Shipment as RequestShipment;
use App\UseCases\Api\HttpRequestExceptionHandler;

class ValidateShipment extends HttpRequestExceptionHandler{

    use Request;

    private const API_URL = '/ship/v1/shipments/packages/validate';
    private const CODE_FOR_VALID_SHIPMENT = 'SHIPMENT.VALIDATION.SUCCESS';

    public function __construct(private readonly Client $client)
    {

    }
     /**
     * Solicitar un envío. FedEx recomienda primero validar que las direcciones y estructura de la solicitud sean válidas,
     * así que primero lo hacemos y luego solicitamos el envío (De momento no tenemos convenio para envíos LTL).
     *
     * @param RequestShipment $request Objeto que contiene los detalles del envío a validar.
     * @return bool Retorna true si la validación es exitosa, de lo contrario lanza una excepción.
     * @throws Fedex Si ocurre un error al conectarse con el servidor o si la estructura del envío es inválida.
     */
    public function handle(RequestShipment $request): bool{

        $response = $this->sendRequest($request);

        $bodyResult = $this->decodeResponse($response);

        return $this->processResponse($bodyResult);

    }
    /**
     * Envía una solicitud de validación de envío a la API de FedEx.
     *
     * @param RequestShipment $request Objeto que contiene los detalles del envío a validar.
     * @return \Psr\Http\Message\ResponseInterface Respuesta de la API de FedEx.
     * @throws Fedex Si no es posible conectarse a la paquetería.
     */
    private function sendRequest(RequestShipment $request): \Psr\Http\Message\ResponseInterface{
        $headers = Factory::credentials();

        try {
            $response = $this->client->post(self::API_URL, [
                'headers' =>[
                    'Authorization' => 'Bearer ' . $headers->authorization,
                    'X-locale' => $headers->xLocale,
                    'Content-Type' => $headers->contentType
                ],
                'json' => $request->array()
            ]);

            if($response->getStatusCode() !== 200) throw new Fedex('La respuesta de la paqueteria es inconsistente ' . json_encode($response), 1);

            return $response;

        }
        catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'ValidateShipment', Fedex::class, $request->array());
        }
    }
    /**
     * Decodifica la respuesta JSON de la API de FedEx.
     *
     * @param \Psr\Http\Message\ResponseInterface $response Respuesta de la API de FedEx.
     * @return object Objeto decodificado de la respuesta JSON.
     * @throws Fedex Si ocurre un error al decodificar la respuesta JSON.
     */
    private function decodeResponse(\Psr\Http\Message\ResponseInterface $response): object {
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Fedex('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    /**
     * Procesa la respuesta de la API de FedEx.
     *
     * @param object $bodyResult Objeto decodificado de la respuesta JSON.
     * @return bool Retorna true si la validación es exitosa.
     * @throws Fedex Si la estructura del envío es inválida.
     */
    private function processResponse(object $bodyResult): bool{
        $messageError = [];

        foreach ($bodyResult->output->alerts as $alert) {

            if($alert->code == self::CODE_FOR_VALID_SHIPMENT) return true;

            $messageError[] = $alert->message;
        }

        throw new Fedex('Estructura del envio invalida: '. implode(',', $messageError));
    }
}
