<?php
namespace App\Services\Paquetexpress;

use App\Services\Paquetexpress\Exceptions\Paquetexpress;
use GuzzleHttp\Client;
use Illuminate\Support\Facades\Cache;
use App\Services\Paquetexpress\Helpers\Request;
use App\Services\Paquetexpress\Objects\User;
use App\Services\Paquetexpress\Objects\Quote\Security;
use App\UseCases\Api\HttpRequestExceptionHandler;
use Illuminate\Support\Facades\Log;

class Credentials extends HttpRequestExceptionHandler{
    use Request;

    private const CACHE_TOKEN = 'api-paquetexpress-token';
    private const API_URL = '/RadRestFul/api/rad/loginv1/login';
    /**
     * El tiempo de vida de un token son 30seg pero le dejamos un margen de 20
     */
    private const TIME_EXPIRE = 20;
    /**
     * Máximo número de intentos de reintento en caso de fallo
     */
    const MAX_RETRIES = 2;

    public function __construct(
        private User $user,
        private readonly Client $client,
        private readonly bool $refresh = false)
    {

    }
    /**
     * Credenciales de acceso, las buscamos en la cache o las solicitamos directamente de la
     * paqueteria
     *
     * @return Security
     */
    public function get(): Security{

        if($this->refresh) $this->getInApi();

        $token = $this->getInCache();
        return $token ? new Security($this->user, $token, 0) : $this->getInApi();

    }
    private function getInApi(): Security {
        $attempts = 0;
        // Lógica de reintento
        while ($attempts < self::MAX_RETRIES) {
            try {

                $response = $this->sendRequest();
                $bodyResult = $this->decodeResponse($response);
                return $this->processResponse($bodyResult);

            } catch (\Throwable $e) {

                $attempts++;

                if ($attempts >= self::MAX_RETRIES) {
                    throw new \Exception("Máximo de errores alcanzados", 1);

                } else {
                    Log::warning("Reintentando solicitud de token a la API de Paquetexpress. Intento: $attempts");

                    sleep(pow(2, $attempts)); // Retraso exponencial antes de reintentar
                }
            }
        }
    }
    private function sendRequest(): \Psr\Http\Message\ResponseInterface{
        try {
            $response = $this->client->post(self::API_URL,
                    [
                        'json' => [
                            'header' => [
                                'security' => $this->user->array()
                            ]
                        ]
                    ]
                );

            if($response->getStatusCode() !== 200) throw new Paquetexpress('No es posible conectarse a la paqueteria ' . json_encode($response), 1);

            return $response;

        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Credentials', Paquetexpress::class);
        }
    }
    private function decodeResponse($response): object{
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Paquetexpress('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    private function processResponse($bodyResult): Security{
        if($bodyResult->body->response->success) {

            $this->saveInCache($bodyResult->body->response->data->token);
            return new Security($this->user, $bodyResult->body->response->data->token, 0);
        }

        $this->errorResponse($bodyResult->body->response);
    }
    private function getInCache():?string {
        return Cache::get(self::CACHE_TOKEN);
    }
    private function saveInCache(string $token): void {
        Cache::set(self::CACHE_TOKEN, $token, now()->addSeconds(self::TIME_EXPIRE));
    }

}
