<?php
namespace App\Services\Paquetexpress;

use App\Models\OrderItem;
use App\Models\Shipment as ModelsShipment;
use App\Services\Paquetexpress\Objects\Address;
use App\Services\Paquetexpress\Objects\Code;
use App\Services\Paquetexpress\Objects\Quote\Device;
use App\Services\Paquetexpress\Objects\Quote\Security;
use App\Services\Paquetexpress\Objects\Quote\Services;
use App\Services\Paquetexpress\Objects\Quote\ServicesAckType;
use App\Services\Paquetexpress\Objects\Quote\ServicesDlvyType;
use App\Services\Paquetexpress\Objects\Quote\ServicesInvType;
use App\Services\Paquetexpress\Objects\Quote\ServicesRadType;
use App\Services\Paquetexpress\Objects\Quote\Target;
use App\Services\Paquetexpress\Objects\QuoteServices;
use App\Services\Paquetexpress\Objects\Shipment;
use App\Services\Paquetexpress\Objects\ShipmentDetails;
use App\Services\Paquetexpress\Objects\User;
use GuzzleHttp\Client;
use App\Services\Paquetexpress\Objects\Shipment\Address as ShipmentAddress;
use App\Services\Paquetexpress\Objects\Shipment\BillRad;
use App\Services\Paquetexpress\Objects\Shipment\ListSrvcItemDTO;
use App\Services\Paquetexpress\Objects\Shipment\PackageItem;
use App\Services\Paquetexpress\Objects\Shipment\RadGuiaAddrDTOList;
use App\Services\Paquetexpress\Objects\Shipment\RadSrvcItemDTOList;
use App\Services\Paquetexpress\Objects\Shipment\SrvcId;
use App\Services\Paquetexpress\Objects\Shipment\SrvcItem;
use App\Services\Paquetexpress\Objects\Shipment\TypeSrvcId;
use App\Services\Paquetexpress\Request\CancelShipment;
use App\Services\Paquetexpress\Request\Quotation;
use App\Services\Paquetexpress\Request\Shipment as RequestShipment;
use App\Services\Shipments\Address as ShipmentsAddress;
use App\Services\Shipments\Contracts\Packaging;
use App\Services\Shipments\Contracts\Product;
use App\Services\Shipments\PerSend\ListPackaging;
use App\Services\Shipments\ShippingQuote;
/**
 * Esta clase nos construye todos los objetos request necesarios para hacer uso de la api
 */
class Factory{
    /**
     * Obtener credenciales de acceso
     * -user
     * -token
     *
     * @param Client $client
     * @return Security
     */
    public static function credentials(Client $client, bool $refresh = false): Security{
        $credentials = new Credentials(
                            new User(
                                config('paquetexpress.user'),
                                config('paquetexpress.password')
                            ),
                            $client,
                            $refresh
                        );
        return $credentials->get();
    }
    /**
     * El acceso para la cotizacion requiere sus propias credenciales
     *
     * @return Security
     */
    public static function quoteCredentials(): Security{
        return new Security(
            new User(
                    config('paquetexpress.user'),
                    config('paquetexpress.password_quotations')
            ),
            config('paquetexpress.token')
        );
    }
    public static function device(): Device{
        return new Device(
            'Customer',
            'Web'
        );
    }
    public static function target(): Target{
        return new Target(
            'QUOTER',
            '1.0',
            'quoter',
            'quotes',
            'R'
        );
    }
    public static function address(
        int $zip_code,
        string $colony,
        string $city = '',
        string $state = '',
        string $country = ''): Address{

        return new Address(
            $zip_code,
            $colony,
            $city,
            $state,
            $country
        );
    }
    /**
     * Servicios extras solicitados, por ejemplo recoleccion del paquete a domicilio y seguro
     *
     * @return Services
     */
    public static function services(): Services{
        return new Services(
            new ServicesDlvyType(ServicesDlvyType::A_DOMICILIO),
            new ServicesAckType(ServicesAckType::SIN_ACUSE),
            new ServicesInvType(ServicesInvType::SIN_SEGURO),
            new ServicesRadType(ServicesRadType::SI)
        );
    }
    public static function quoteServices(): QuoteServices{
        return new QuoteServices([QuoteServices::ALL]);
    }
    public static function service(): Service{
        return new Service(new Client(
            ['base_uri' => config('paquetexpress.base_uri')]
        ));
    }
    public static function quote(ShipmentsAddress $origin, ShippingQuote $shipment): Quotation{

        return new Quotation(
                Factory::quoteCredentials(),
                Factory::device(),
                Factory::target(),
                Factory::address($origin->cp, $origin->colonyName),
                Factory::address($shipment->destination->cp, $shipment->destination->colonyName),
                Factory::quotationDetailByShipment($shipment),
                Factory::services(),
                Factory::quoteServices()
            );
    }
    public static function quotationDetailByShipment(ShippingQuote $shipment): ShipmentDetails{
        return new ShipmentDetails(self::shipmentItems($shipment->getListPackaging()));
    }
    private static function shipmentItems(ListPackaging $listPackaging){
        $items = $listPackaging->packages();

        $callback = function(Packaging $item, int $counter){
            $weight = $item->getWeight(1000);
            $length = $item->getLength(10);
            $width = $item->getWidth(10);
            $depth = $item->getDepth(10);
            //Cuando cotizamos siempre indicamos que el envio es una caja, cuando se solicita la guia ya no es asi
            return new Shipment(
                $counter,
                1,
                new Code(Code::CAJA),
                $weight,
                $length,
                $width,
                $depth
            );
        };

        $shipmentItems = [];

        foreach ($items as $key => $item) {
            $shipmentItems[] = $callback($item, $key + 1);
        }
        return $shipmentItems;
    }
    /**
     *
     *
     * @param [type] $shipmentId
     * @param ListPackaging  $listPackaging
     * @return RequestShipment
     */
    public static function shipmentData($shipmentId, ListPackaging $listPackaging, Security $security): RequestShipment{
        $shipment = ModelsShipment::forShipping($shipmentId);

        return new RequestShipment(
                $security,
                self::device(),
                self::target(),
                new BillRad(BillRad::REQUEST),
                config('paquetexpress.billClntId'),
                self::radGuiaAddrDTOList($shipment),
                new RadSrvcItemDTOList(self::radSrvcItemDTOList($listPackaging)),
                new ListSrvcItemDTO([
                    new SrvcItem('RAD'),
                    new SrvcItem('EAD')
                ]),
                new TypeSrvcId($shipment->service_code)
            );
    }
    public static function cancelShipment(string $trackingNumber, Security $security): CancelShipment{
        return new CancelShipment($security, $trackingNumber);
    }
    private static function radGuiaAddrDTOList($shipment): RadGuiaAddrDTOList{
        return new RadGuiaAddrDTOList(
            new ShipmentAddress(
                $shipment->origin_country,
                $shipment->origin_state,
                $shipment->origin_city,
                $shipment->origin_city,
                $shipment->origin_neighborhood,
                $shipment->origin_postal_code,
                $shipment->origin_address,
                $shipment->origin_contact_phone,
                $shipment->origin_contact_name,
                '.',
                $shipment->origin_contact_name,
                'ORIGIN',
                $shipment->getFullOriginNumber()
            ),
            new ShipmentAddress(
                $shipment->destination_country,
                $shipment->destination_state,
                $shipment->destination_city,
                $shipment->destination_city,
                $shipment->destination_neighborhood,
                $shipment->destination_postal_code,
                $shipment->destination_address,
                $shipment->destination_contact_phone,
                $shipment->destination_contact_name,
                '.',
                $shipment->destination_contact_name,
                'DESTINATION',
                $shipment->getFullDestinationNumber()
            )
        );
    }
    /**
     *De acuerdo con las directrices establecidas por Paquetexpress en las comunicaciones electrónicas,
     *es necesario seleccionar el producto con el valor más alto del conjunto de productos enviados,
     *esta informacion sera la que aparezca como el contenido del paquete ademas de que sera la que se
     *reportara al SAT
     *Ver http://omawww.sat.gob.mx/tramitesyservicios/Paginas/complemento_carta_porte.htm
     *
     * @param ListPackaging  $listPackaging
     * @return array
     */
    private static function radSrvcItemDTOList(ListPackaging $listPackaging){

        return array_map(function(\App\Services\Shipments\Contracts\Packaging $packaging){

                $itemMostExpense = array_reduce($packaging->getItems(), function(Product|null $itemMostExpense, Product $orderItem){

                    if($itemMostExpense === null || $orderItem->subtotal() > $itemMostExpense->subtotal())
                        $itemMostExpense = $orderItem;

                    return $itemMostExpense;
                });

                $weight = $packaging->getWeight(1000);
                $length = $packaging->getLength(10);
                $width = $packaging->getWidth(10);
                $depth = $packaging->getDepth(10);

                return new PackageItem(
                    SrvcId::equivalent($weight, $length, $width, $depth)->value,
                    $itemMostExpense->claveProductoServicioCP(),
                    $packaging->getWeight(1000),
                    $packaging->getLength(10),
                    $packaging->getWidth(10),
                    $packaging->getDepth(10),
                    $itemMostExpense->getDescription(),
                    1
                );
            }, $listPackaging->packages());
    }
}
