<?php

namespace App\Services\Paquetexpress;

use App\Services\Paquetexpress\Exceptions\Paquetexpress;
use GuzzleHttp\Client;
use App\Services\Paquetexpress\Helpers\Request;
use App\Services\Paquetexpress\Request\Quotation;
use App\Services\Paquetexpress\Response\Quotation as ResponseQuotation;
use App\Services\Shipments\ShippingQuote;
use App\UseCases\Api\HttpRequestExceptionHandler;

class Quote extends HttpRequestExceptionHandler{

    use Request;

    private const API_URL = '/WsQuotePaquetexpress/api/apiQuoter/v2/getQuotation';

    public function __construct(public readonly Client $client)
    {

    }
    /**
     * Cotizar N envios, de manera predeterminada solicitamos todos los tipos de envios
     *
     * @param Quotation $data
     * @return ResponseQuotation
     */
    public function handle(ShippingQuote $shipment){
        $quotes = [];

        foreach ($shipment->addresses as $address) {

            $quote = Factory::quote($address, $shipment);

            try {
                \Log::info(json_encode($quote->array()));
                $response = $this->sendRequest($quote);

            } catch (\Throwable $th) {
                $this->errorResponse($th->getMessage(), false);
                continue;
            }

            $bodyResult = $this->decodeResponse($response);
            \Log::info(json_encode($bodyResult));
            if ($bodyResult->body->response->success) {
                $quotes[] = new ResponseQuotation($bodyResult->body->response->data, $address, $shipment->destination);
            }else
                $this->errorResponse($bodyResult->body->response, false);
        }

        if(empty($quotes)) throw new Paquetexpress('Sin ninguna cotizacion de PaquetExpress', 1);

        $quote = count($quotes) == 1 ? $quotes[0] : $this->selectBestQuote($quotes);

        return $quote->array();
    }
    /**
     * Envía la solicitud de cotización a la API.
     *
     * @param Quotation $quote
     * @return \Psr\Http\Message\ResponseInterface
     * @throws Paquetexpress
     */
    private function sendRequest(Quotation $quote){
        try {

            $response = $this->client->post(self::API_URL, ['json' => $quote->array()]);

            if($response->getStatusCode() !== 200) throw new Paquetexpress('No es posible conectarse a la paqueteria ' . json_encode($response), 1);

            return $response;

        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Quote', Paquetexpress::class, $quote->array());
        }
    }

    private function decodeResponse($response) {
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Paquetexpress('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }

    private function selectBestQuote($quotations)
    {
        $cheapestQuotation = array_pop($quotations);

        $cheapestQuotation = array_reduce($quotations,  function(ResponseQuotation $cheapestQuotation, ResponseQuotation $quote){

            if($quote->cost() < $cheapestQuotation->cost()) $cheapestQuotation = $quote;

            return $cheapestQuotation;
        }, $cheapestQuotation);

        return $cheapestQuotation;
    }
}
