<?php

namespace App\Services\Paquetexpress;

use App\Services\Paquetexpress\Exceptions\Paquetexpress;
use GuzzleHttp\Client;
use App\Services\Paquetexpress\Helpers\Request;
use App\Services\Paquetexpress\Response\Shipment as ResponseShipment;
use App\Services\Shipments\Packer;
use App\Services\Paquetexpress\Request\Shipment as RequestShipment;
use App\Services\Shipments\PerSend\ListPackaging;
use Illuminate\Support\Facades\Log;
use App\UseCases\Api\HttpRequestExceptionHandler;

class Shipment extends HttpRequestExceptionHandler{

    use Request;
    private const MAX_INTENTS = 2;
    /**
     * Contador de intentos para reintentar por un error en la autenticacion
     *
     * @var integer
     */
    private  int $intents = 0;

    private const API_URL = '/RadRestFul/api/rad/v1/guia';

    public function __construct(public readonly Client $client)
    {

    }
    /**
     * Solicitar una guia y guardarla
     *
     * @param int $shipmentId
     * @return ResponseShipment
     */
    public function handle(int $shipmentId){
        $listPackaging = Packer::byShipment($shipmentId);
        $data = Factory::shipmentData($shipmentId, $listPackaging, Factory::credentials($this->client, $this->intents > 0));
        \Log::info(json_encode($data->array()));
        $response = $this->sendRequest($data);

        $bodyResult = $this->decodeResponse($response);

        $response = $this->processResponse($bodyResult, $listPackaging, $shipmentId);

        $shipmentLabel = new ShipmentLabel($this->client);

        if(!$shipmentLabel->handle($shipmentId, $response->getTrackingNumber()))
            Log::error('Hubo un error al descargar el PDF de la guía: ' . $shipmentId);

        return $response;
    }

    private function processResponse(object $bodyResult, ListPackaging  $listPackaging, int $shipmentId): ResponseShipment
    {
        $responseBody = $bodyResult->body->response ?? null;

        if($responseBody && $responseBody->success){
            return new ResponseShipment($bodyResult,  $listPackaging, $shipmentId);
        };
        //Buscamos si hay un problema con la autentificacion
        if($responseBody
            && !$responseBody->success
            && isset($responseBody->messages)
            && is_array($responseBody->messages)
            && count($responseBody->messages) > 0){
            if($responseBody->messages[0]->description == "The Token's Signature resulted invalid when verified")
                $this->reintentAuthAndRequest($shipmentId);
        }

        $this->errorResponse($bodyResult->body->response);
    }
    private function sendRequest(RequestShipment $data){
        try {

            $response = $this->client->post(self::API_URL, ['json' => $data->array()]);

            if($response->getStatusCode() !== 200) throw new Paquetexpress('No es posible conectarse a la paqueteria ' . json_encode($response), 1);

            return $response;

        }
        catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Shipment', Paquetexpress::class, $data->array());
        }
    }

    private function decodeResponse(\Psr\Http\Message\ResponseInterface $response): object {
        $bodyResult = json_decode($response->getBody());
        Log::info(json_encode($bodyResult));
        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Paquetexpress('Error al decodificar la respuesta JSON, respuesta: ' . $response->getBody());
        }

        return $bodyResult;
    }
    private function reintentAuthAndRequest(int $shipmentId){
        $this->intents++;
        Log::info("Paquetexpress: Reintentando autenticacion, intentos ($this->intents)");
        $this->handle($shipmentId);
    }
}
