<?php

namespace App\Services\Paquetexpress;

use App\Services\Paquetexpress\Exceptions\Paquetexpress;
use GuzzleHttp\Client;
use App\Services\Paquetexpress\Helpers\Request;
use App\Services\Paquetexpress\Response\Tracking as ResponseTracking;
use App\UseCases\Api\HttpRequestExceptionHandler;

class Tracking extends HttpRequestExceptionHandler{

    use Request;

    public function __construct(public readonly Client $client)
    {

    }
     /**
     * Maneja la solicitud de rastreo de un envío.
     *
     * @param string $trackingId ID de rastreo del envío.
     * @return ResponseTracking Objeto que contiene la respuesta del rastreo.
     * @throws Paquetexpress Si ocurre un error al conectarse con la paquetería o al decodificar la respuesta JSON.
     */
    public function handle(string $trackingId): ResponseTracking{

        $response = $this->sendRequest($trackingId);

        $bodyResult = $this->decodeResponse($response);

        return $this->processResponse($bodyResult);

    }
    /**
     * Envía una solicitud de rastreo a la API de Paquetexpress.
     *
     * @param string $trackingId ID de rastreo del envío.
     * @return \Psr\Http\Message\ResponseInterface Respuesta de la API de Paquetexpress.
     * @throws Paquetexpress Si no es posible conectarse a la paquetería.
     */
    private function sendRequest(string $trackingId): \Psr\Http\Message\ResponseInterface{
        try {
            $token = config('paquetexpress.token');
            $response = $this->client->get(config('paquetexpress.tracking_url') . "$trackingId/$token");

            if($response->getStatusCode() !== 200) throw new Paquetexpress('No es posible conectarse a la paqueteria: ' . json_encode($response), 1);

            return $response;
        } catch (\Throwable $e) {
            $this->handleHttpRequestException($e, 'Tracking', Paquetexpress::class, ['url' => config('paquetexpress.tracking_url') . "$trackingId/$token"]);
        }
    }
    /**
     * Decodifica la respuesta JSON de la API de Paquetexpress.
     *
     * @param \Psr\Http\Message\ResponseInterface $response Respuesta de la API de Paquetexpress.
     * @return object Objeto decodificado de la respuesta JSON.
     * @throws Paquetexpress Si ocurre un error al decodificar la respuesta JSON.
     */
    private function decodeResponse($response): object{
        $bodyResult = json_decode($response->getBody());

        if (json_last_error() !== JSON_ERROR_NONE) {
            throw new Paquetexpress('Error al decodificar la respuesta JSON');
        }

        return $bodyResult;
    }
    /**
     * Procesa la respuesta de la API de Paquetexpress.
     *
     * @param object $bodyResult Objeto decodificado de la respuesta JSON.
     * @return ResponseTracking Objeto que contiene la respuesta del rastreo.
     * @throws Paquetexpress Si la estructura del envío es inválida.
     */
    private function processResponse(object $bodyResult): ResponseTracking{

        if($bodyResult->body->response->success && count($bodyResult->body->response->data) > 0)
            return new ResponseTracking($bodyResult);

        $this->errorResponse($bodyResult->body->response);
    }
}
