<?php
namespace App\Services\Shipments\Packaging;

use App\Services\Shipments\Contracts\Product;
use DVDoug\BoxPacker\Box;
use DVDoug\BoxPacker\PackedItem;
use DVDoug\BoxPacker\PackedItemList;
use DVDoug\BoxPacker\ConstrainedPlacementItem;

class Panel implements Product, ConstrainedPlacementItem{

    public function __construct(
        private readonly Product $product)
    {

    }
    /**
     * Subtotal del costo de este producto resultado de multiplicar cantidad * costo por unidad
     *
     * @return float
     */
    public function subtotal(): float{
        return $this->product->subtotal();
    }
    /**
     * Cantidad de unidades del producto
     *
     * @return float
     */
    public function quantity(): float{
        return $this->product->quantity();
    }
    /**
     * Identificador unico del producto, por ejemplo el ID en la BD
     *
     * @return integer
     */
    public function identification(): int{
        return $this->product->identification();
    }
    /**
     * Costo del producto
     *
     * @return float
     */
    public function cost(): float{
        return $this->product->cost();
    }
    /**
     * Item SKU etc.
     */
    public function getDescription(): string{
        return $this->product->getDescription();
    }

    /**
     * Item width in mm.
     */
    public function getWidth(): int{
        return $this->product->getWidth();
    }

    /**
     * Item length in mm.
     */
    public function getLength(): int{
        return $this->product->getLength();
    }

    /**
     * Item depth in mm.
     */
    public function getDepth(): int{
        return $this->product->getDepth();
    }

    /**
     * Item weight in g.
     */
    public function getWeight(): int{
        return $this->product->getWeight();
    }

    /**
     * Does this item need to be kept flat / packed "this way up"?
     */
    public function getKeepFlat(): bool{
        return $this->product->getKeepFlat();
    }
    public function rotate90Y(): void{
        $this->product->rotate90Y();
    }
    /**
     * Indicar si el producto puede ser rotado
     *
     * @param integer $rule
     * @return void
     */
    public function setKeepFlat(int $rule): void{
        $this->product->setKeepFlat($rule);
    }
    public function isRotateIn90Y(): bool{
        return $this->product->isRotateIn90Y();
    }
    public function claveProductoServicioCP(): string{
        return $this->product->claveProductoServicioCP();
    }
    public function claveUnidadPeso(): string{
        return $this->product->claveUnidadPeso();
    }
    /**
     * Si el panel fue rotado no puede llevar nada encima.
     *
     * @param  Box            $box
     * @param  PackedItemList $alreadyPackedItems
     * @param  int            $proposedX
     * @param  int            $proposedY
     * @param  int            $proposedZ
     * @param  int            $width
     * @param  int            $length
     * @param  int            $depth
     * @return bool
     */
    public function canBePacked(
        Box $box,
        PackedItemList $alreadyPackedItems,
        int $proposedX,
        int $proposedY,
        int $proposedZ,
        int $width,
        int $length,
        int $depth
    ): bool {

        /** @var PackedItem $alreadyPacked */
        foreach ($alreadyPackedItems as $alreadyPacked) {
            if($this->inSamePosition($alreadyPacked, $proposedZ, $proposedX, $proposedY)){

                if(!$this->isRotateIn90Y() && $alreadyPacked->getItem()->isRotateIn90Y()){
                    return false;
                }

                if($this->isRotateIn90Y()){
                    return false;
                }

            }
        }

        return true;
    }

    private function inSamePosition(PackedItem $alreadyPacked, $proposedZ, $proposedX, $proposedY){

        return $alreadyPacked->getZ() + $alreadyPacked->getDepth() === $proposedZ &&
                $proposedX >= $alreadyPacked->getX() &&
                $proposedX <= ($alreadyPacked->getX() + $alreadyPacked->getWidth()) &&
                $proposedY >= $alreadyPacked->getY() &&
                $proposedY <= ($alreadyPacked->getY() + $alreadyPacked->getLength());
    }
}
