<?php
namespace App\Services\Shipments\Packaging;

use App\Services\Shipments\Contracts\Product AS Contract;
use App\Services\Shipments\Product;
use DVDoug\BoxPacker\Rotation;
use DVDoug\BoxPacker\ItemList;
/**
 * Esta clase identifica si un producto que será embalado es un panel solar, ademas de expecificar reglas de operaciones
 * y su aplicacion antes de embalar
 */
class Panels{
    private array $panels = [];
    /**
     * Aqui guardaremos los paneles que rotaremos, esta variable se resetea cuando se alcanza el limite de paneles
     * solares por tarima
     *
     * @var array
     */
    private array $forRotate = [];
    /**
     * Aqui guardamos el ancho que se necesitaria para los paneles que se van a rotar
     *
     * @var float
     */
    private float $widthUsed = 0;
    /**
     * Ancho maximo de la tarima mas grande, lo usamos como limite para saber cuantos paneles en posicion vertical le caben
     */
    public const PALLET_WIDTH_AVAILABLE = 1140;
    /**
     * Maximo de paneles que pueden ir apilados
     */
    public const MAX_STACKABLE_PANELS = 10;
    /**
     * Minimo de espacio requerido para un panel
     */
    public const MIN_SPACE_FOR_PANEL = 30;
    /**
     * Longitud minima de un panel solar (mm)
     */
    private const MIN_LENGTH = 1600;
    /**
     * Ancho minimo de un panel solar (mm)
     */
    private const MIN_WIDTH = 1000;
    /**
     * Alto minimo de un panel solar (mm)
     */
    private const MIN_DEPTH = 20;
    /**
     * Longitud maxima de un panel solar (mm)
     */
    private const MAX_LENGTH = 2500;
    /**
     * Ancho maximo de un panel solar (mm)
     */
    private const MAX_WIDTH = 1400;
    /**
     * Alto maximo de un panel solar (mm)
     */
    private const MAX_DEPTH = 40;
    /**
     * Identificar si un producto es un panel solar (El producto no debe haber sido rotado previamente)
     *
     * @param Contract $product
     * @return boolean
     */
    public static function isPanel(Contract $product): bool{

        return $product->getLength() >= self::MIN_LENGTH &&
                $product->getLength() <= self::MAX_LENGTH &&
                $product->getWidth() >= self::MIN_WIDTH &&
                $product->getWidth() <= self::MAX_WIDTH &&
                $product->getDepth() >= self::MIN_DEPTH &&
                $product->getDepth() <= self::MAX_DEPTH;
    }
    public function addProduct(Product $product){
        //CLonamos el producto porque lo vamos a alterar si es necesario rotarlo
        for ($i = 0; $i < $product->quantity() ; $i++) {
            $this->panels[] = new Panel(clone $product);
        }
    }
    public function getPanelList(): ItemList{
        $this->rotatePanels();
        $itemList = new ItemList();

        foreach ($this->panels as $panel) {
            $itemList->insert($panel, 1);
        }
        return $itemList;
    }
    /**
     * Rotar los paneles en el eje Y si se cumplen las siguientes condiciones
     * Cuando la cantidad de paneles supere el valor de 11 Paneles
     *
     * @return void
     */
    private function rotatePanels(){

        array_walk($this->panels, function(Panel $product){
            $totalPanels = count($this->forRotate);

            if($totalPanels > Panels::MAX_STACKABLE_PANELS && $this->widthLimitReached()){
                array_walk($this->forRotate, function(Panel $panel){
                    $panel->rotate90Y();
                    $panel->setKeepFlat(Rotation::KeepFlat);
                });
                $this->forRotate = [];
                $this->widthUsed = 0;
            }

            $this->forRotate[] = $product;
            $this->widthUsed += $product->getDepth();
        });
        //Aqui rotamos los paneles restantes, en caso de ser necesario
        if(count($this->forRotate) > Panels::MAX_STACKABLE_PANELS){
            array_walk($this->forRotate, function(Panel $panel){
                $panel->rotate90Y();
                $panel->setKeepFlat(Rotation::KeepFlat);
            });
        }

    }
    /**
     * Se alcanzo el limite de ancho disponible de la tarima mas grande que se usa
     *
     * @return bool
     */
    private function widthLimitReached(){
        return ($this->widthUsed + self::MIN_SPACE_FOR_PANEL) > self::PALLET_WIDTH_AVAILABLE;
    }
}
