<?php
namespace App\Services\Shipments;

use App\Exceptions\Handler;
use App\Models\ShippingBox;
use App\Services\Shipments\PerSend\ListPackaging;
use App\Services\Shipments\PerSend\RegularContainer;
use App\Services\Shipments\PerSend\SelfContainer;
use DVDoug\BoxPacker\BoxList;
use DVDoug\BoxPacker\ItemList;
use DVDoug\BoxPacker\NoBoxesAvailableException;
use DVDoug\BoxPacker\Packer as BoxPackerPacker;
use Illuminate\Support\Collection;
use DVDoug\BoxPacker\InfalliblePacker;

/**
 * Esta clase empaqueta un envio
 */
class Packer {
    /**
     *
     *
     * @param Collection $shipmentBoxes Todas las cajas donde puede ser enviada una compra
     * @param PackagedShipping $package Conjunto de productos que seran enviados
     */
    public function __construct(private Collection $shipmentBoxes, private PackagedShipping $package)
    {

    }

    public function handle(): ListPackaging{

        $listPackaging = new ListPackaging();

        $this->package->handle();

        $this->packPanels($listPackaging);
        $this->packOthers($listPackaging);

        return $listPackaging;
    }

    /**
     *
     *
     * @return ListPackaging
     */
    public static function calculatePackaging(Collection $shipmentBoxes, PackagedShipping $package): ListPackaging{
        $handler = new self($shipmentBoxes, $package);

        return $handler->handle();
    }
    /**
     * Obtener el empaquetado de un envio ya guardado
     *
     * @param integer $shipmentId
     * @return ListPackaging
     */
    public static function byShipment(int $shipmentId): ListPackaging{
        return Packer::calculatePackaging(ShippingBox::boxList(), PackagedShipping::byShipment($shipmentId));
    }
    // Si el envio es en pallet se usa el empaquetador normal, caso contrario usamos el que permite items sin
    // * empaquetar
    // * Esta parte es fundamental, si el envio no es una pallet permitiremos items que no sean empaquetados,
    // * y estos seran considerados como self-packagin
    private function pack(
        ListPackaging $listPackaging,
        ItemList $list,
        Collection $shipmentBoxes,
        InfalliblePacker|BoxPackerPacker $packer,
        TypeOfPackaging $typeOfPackaging ): ListPackaging{
        try {

            $packer->setBoxes((new BoxList())->fromArray($shipmentBoxes->all()));

            $packer->setItems($list);

            $packedBoxes = $packer->pack();

            foreach ($packedBoxes as $packedBox) {
                $shippingBox = $packedBox->getBox();
                $listPackaging->addPackage(new RegularContainer($shippingBox, $packedBox, $typeOfPackaging));
            }

            if($typeOfPackaging === TypeOfPackaging::BOX) {
                $unpackedItems = $packer->getUnpackedItems();

                foreach ($unpackedItems as $item) {
                    $listPackaging->addUnpackage(new SelfContainer($item));
                }

            }
            return $listPackaging;
        } catch (NoBoxesAvailableException $th) {
            Handler::log($th);
            throw new \Exception("Tu pedido es demasiado grande, contacta a ventas por favor", 1);
        }

    }
    /**
     * Los paneles solares siempre requieren tarima y no se combinan con otros paquetes
     *
     * @param ListPackaging $listPackaging
     * @return void
     */
    private function packPanels(ListPackaging $listPackaging){
        $listPanels = $this->package->getPanelList();

        if($listPanels->count() > 0)
            $this->pack(
                $listPackaging,
                $listPanels,
                $this->shipmentBoxes->filter(fn(ShippingBox $shipmentBox) => $shipmentBox->type == TypeOfPackaging::PALLET->value),
                new BoxPackerPacker(),
                TypeOfPackaging::PALLET
            );
    }
    private function packOthers(ListPackaging $listPackaging){
        $listOthers = $this->package->getOthersList();
        $typeOfPackaging = TypeOfPackaging::fromItems($listOthers);

        if($listOthers->count() > 0)
            $this->pack(
                $listPackaging,
                $listOthers,
                $this->shipmentBoxes->filter(fn(ShippingBox $shipmentBox) => $shipmentBox->type == $typeOfPackaging->value),
                new InfalliblePacker(),
                $typeOfPackaging
            );
    }
}
