<?php
namespace App\Services\Shipments\PerSend;

use App\Models\ShippingBox;
use App\Services\Helpers\Money;
use App\Services\Shipments\Contracts\Packaging as ContractsPackaging;
use App\Services\Shipments\Contracts\Product;
use App\Services\Shipments\TypeOfPackaging;
use DVDoug\BoxPacker\PackedBox;
/**
 * Esta clase representa un contenedor por enviar
 */
class RegularContainer implements ContractsPackaging{

    /**
     *
     *
     * @param ShippingBox $shippingBox
     * @param float $weight in g
     */
    public function __construct(
        public readonly ShippingBox $shippingBox,
        public readonly PackedBox $packedBox,
        public readonly TypeOfPackaging $type
    ) {
    }
 /**
     * Get packed weight.
     * @param $divisor Si te interesa que el valor sea dividido indica el divisor
     * @return int weight in grams (or divided)
     */
    public function getWeight($divisor = 1): float{
        return $this->packedBox->getWeight() / $divisor;
    }
    /**
     * Length of package in mm..
     * @param $divisor Si te interesa que el valor sea dividido indica el divisor
     */
    public function getLength($divisor = 1): float{
        return $this->shippingBox->getOuterLength() / $divisor;
    }
    /**
     * Width of package in mm.
     * @param $divisor Si te interesa que el valor sea dividido indica el divisor
     */
    public function getWidth($divisor = 1): float{
        return $this->shippingBox->getOuterWidth() / $divisor;
    }
    /**
     * Depth of package in mm. (When is a pallet, the Depth returned is the Depth was used)
     * @param $divisor Si te interesa que el valor sea dividido indica el divisor
     */
    public function getDepth($divisor = 1): float{

        if($this->type == TypeOfPackaging::PALLET)
            return $this->packedBox->getUsedDepth() / $divisor;

        return $this->shippingBox->getOuterDepth() / $divisor;
    }
    /**
     * Valor total del paquete.
     * Esta funcion solo funciona si los items que fueron ingresados implementan la interfaz Products
     *
     * @return float
     */
    public function getTotal(): float{
        /**
         * getItems nos devuelve todos los items guardados, TODOS, lo que implica que si hay
         * mas de un item del mismo tipo este se repite, hay que filtrar siempre usando
         * el identificador
         */
        $packedItems = $this->packedBox->getItems();
        $total = 0;
        $alreadyUsed = [];
        foreach ($packedItems as $packedItem) {

            $item = $packedItem->getItem();

            if($item instanceof Product && !in_array($item->identification(), $alreadyUsed)){
                $total = Money::add($total, $item->subtotal());
                $alreadyUsed[] = $item->identification();
            }
        }

        return $total;
    }
    /**
     * Obtener los items originales empaquetados
     *
     * @return array
     */
    public function getItems(): array{
        /**
         * getItems nos devuelve todos los items guardados, TODOS, lo que implica que si hay
         * mas de un item del mismo tipo este se repite, hay que filtrar siempre usando
         * el identificador
         */
        $packedItems = $this->packedBox->getItems();
        $items = [];
        $alreadyUsed = [];
        foreach ($packedItems as $packedItem) {

            $item = $packedItem->getItem();

            if($item instanceof Product && !in_array($item->identification(), $alreadyUsed)){
                $items[] = $packedItem->getItem();
                $alreadyUsed[] = $item->identification();
            }

        }

        return $items;
    }
}
