<?php

namespace App\Services\Teams\AdaptiveCards;

use App\Models\Customer;
use App\Models\Order;
use App\Services\Helpers\Money;
use App\Services\Teams\Objects\Product;
use Illuminate\Support\Str;

class OrderPaid
{
    private float $total = 0;

    public function __construct(
        private readonly Customer $customer,
        private readonly Order $order,
        private readonly array $products
    ) {}

    public function build(): array
    {
        return [
            "type" => "AdaptiveCard",
            "attachments" => [
                [
                    "contentType" => "application/vnd.microsoft.card.adaptive",
                    "content" => [
                        '$schema' => "http://adaptivecards.io/schemas/adaptive-card.json",
                        "type" => "AdaptiveCard",
                        "version" => "1.4",
                        "msteams" => [
                            "width" => "Full"
                        ],
                        'channel' => 'b2baa290-1eaf-44cc-8c62-736616d22d8c',
                        "body" => [
                            $this->title(),
                            $this->tableHead(),
                            $this->tableBody(),
                            $this->actionLink()
                        ]
                    ]
                ]
            ]
        ];
    }

    private function title(): array
    {
        return [
            "type" => "ColumnSet",
            "style" => "good",
            "columns" => [
                [
                    "type" => "Column",
                    "width" => "stretch",
                    'items' => [
                        [
                            "type" => "TextBlock",
                            "size" => "Medium",
                            "weight" => "Bolder",
                            "text" => "Cliente " . $this->customer->company_name
                        ]
                    ]
                ],
                [
                    "type" => "Column",
                    "width" => "auto",
                    'items' => [
                        [
                            "type" => "TextBlock",
                            "size" => "Medium",
                            "weight" => "Bolder",
                            "text" => "Venta completada"
                        ]
                    ]
                ],
                [
                    "type" => "Column",
                    "width" => "stretch"
                ]
            ]
        ];
    }

    private function tableHead(): array
    {
        return [
            "type" => "ColumnSet",
            "columns" => [
                $this->tableHeaderColumn("Imagen"),
                $this->tableHeaderColumn("Producto"),
                $this->tableHeaderColumn("Cantidad"),
                // $this->tableHeaderColumn("Precio"),
                // $this->tableHeaderColumn("Subtotal"),
            ],
            "style" => "emphasis"
        ];
    }

    private function tableHeaderColumn(string $headerText): array
    {
        return [
            "type" => "Column",
            "width" => "stretch",
            "items" => [
                [
                    "type" => "TextBlock",
                    "text" => $headerText,
                    "horizontalAlignment" => "Center",
                    "fontType" => "Monospace",
                    "size" => "Medium",
                    "weight" => "Bolder",
                    "color" => "Light"
                ]
            ]
        ];
    }

    private function tableBody(): array
    {
        $container = ["type" => "Container", 'items' => []];

        foreach ($this->products as $product) {
            $container['items'][] = $this->productRow($product);
            $this->total = Money::add($this->total, $product->subtotal);
        }

        $container['items'][] = $this->totalRow();

        return $container;
    }

    private function productRow(Product $product): array
    {
        $url = Str::startsWith($product->imageUrl, 'https://') ? $product->imageUrl : config('notifications.default_img');

        return [
            "type" => "ColumnSet",
            "columns" => [
                $this->productColumnImage($url),
                $this->productColumn($product->name, 'stretch'),
                $this->productColumn($product->quantity, '50px'),
                // $this->productColumn(Money::format($product->price)),
                // $this->productColumn(Money::format($product->subtotal)),
            ],
            'separator' => true
        ];
    }
    private function productColumnImage(string $url){
        return [
            "type" => "Column",
            "width" => "auto",
            "items" => [
                [
                   "type" => "Image",
                    "url" => $url,
                    "size" => "Large",
                ]
            ]
        ];
    }
    private function productColumn(string $text, string $width = 'auto'): array
    {
        return [
            "type" => "Column",
            "width" => $width,
            "items" => [
                [
                    "type" => "TextBlock",
                    "text" => $text,
                    "horizontalAlignment" => "Center",
                    "fontType" => "Monospace"
                ]
            ]
        ];
    }

    private function actionLink(): array
    {
        return [
            "type" => "ColumnSet",
            "columns" => [
                ["type" => "Column", "width" => "stretch"],
                [
                    "type" => "Column",
                    "width" => "auto",
                    "items" => [
                        [
                            "type" => "ActionSet",
                            "actions" => [
                                [
                                    "type" => "Action.OpenUrl",
                                    "title" => "Ver",
                                    "style" => "positive",
                                    'url' => route('admin.orders.edit', ['id' => $this->order->id])
                                ]
                            ]
                        ]
                    ]
                ],
                ["type" => "Column", "width" => "stretch"]
            ]
        ];
    }

    private function totalRow(): array
    {
        return [
            "type" => "ColumnSet",
            "style" => "emphasis",
            "columns" => [
                ["type" => "Column", "width" => "stretch"],
                ["type" => "Column", "width" => "stretch"],
                [
                    "type" => "Column",
                    "width" => "auto",
                    "items" => [
                        [
                            "type" => "TextBlock",
                            "text" => 'Total: ' . Money::format($this->total),
                            "horizontalAlignment" => "Center",
                            "fontType" => "Monospace"
                        ]
                    ]
                ]
            ]
        ];
    }
}
