<?php

namespace App\UseCases\Order;

use App\Enums\OrderStatusEnum;
use App\Exceptions\Handler;
use App\Models\Order;
use App\Models\OrderCanceled;
use App\Models\User;
use App\Notifications\Customer\Order\Canceled;
use Illuminate\Support\Facades\DB;
/**
* La clase Cancel es responsable de manejar la lógica para cancelar un pedido. Cuando se cancela un pedido, la clase:
* 1. Registra la cancelación en la base de datos.
* 2. Cambia el estado del pedido a "Cancelado".
* 3. Notifica al cliente de que el pedido ha sido cancelado.
*/
class Cancel{

    public function __construct(
        public readonly int $userId,
        public readonly Order $order,
        public readonly User $owner,
        private readonly string $reason)
    {

    }
    /**
     * Ejecuta el proceso de cancelación de un pedido.
     * @return bool
     */
    public function handle(){
        try {

            DB::beginTransaction();

            $this->logOfCancellation();

            $this->order->status = OrderStatusEnum::CANCELLED;

            if($this->order->save()){

                $this->restoreStock();

                $this->notifyCustomer();

                DB::commit();
                return true;
            }

        } catch (\Throwable $th) {
            Handler::log($th);
            DB::rollBack();
        }

        throw new \Exception("Hubo un problema al actualizar la orden: $this->order->id ", 1);
    }
    private function notifyCustomer(){
        $this->owner->notify(new Canceled($this->order));
    }
    private function logOfCancellation(){
        OrderCanceled::create([
            'user_id' => $this->userId,
            'order_id' => $this->order->id,
            'reason' => $this->reason
        ]);
    }
    private function restoreStock(){

        $handleRestock = new RestockOrderItems($this->order);
        if(!$handleRestock->handle()) throw new \Exception("Hubo un problema al restablecer el stock de la orden: $this->order->id", 1);

    }
    public static function can(Order $order){
        // Verifica si la orden está en un estado que permite su cancelación
        $isCancellable = match ($order->status) {
            OrderStatusEnum::PAYMENT_SUBMITTED => true,
            OrderStatusEnum::PENDING_PAYMENT => true,
            OrderStatusEnum::PENDING => true,
            OrderStatusEnum::APPROVED => true,
            OrderStatusEnum::WITH_GUIDE => true,
            default => false
        };
         // La orden es cancelable solo si está en un estado permitido y todos los envíos están inactivos
        return $isCancellable && Order::hasAllShipmentsInactives($order->id);
    }
}
