<?php
namespace App\UseCases\Order\Customer\SaveByRequest;

use App\Exceptions\Handler;
use App\Models\Customer;
use App\Models\Order;
use App\Models\OrderItem;
use App\Models\ProductStock;
use App\Models\Shipment;
use App\Services\Cart\CartCondition;
use App\Services\Helpers\Money;
use Darryldecode\Cart\CartCollection;
use App\Services\Cart\Factory as CartConditions;
use App\Services\Cart\Rules;
use App\Services\Teams\Objects\Product as ObjectsProduct;
use Illuminate\Support\Facades\DB;
/**
 * Esta clase guarda una compra de un cliente
 * 1. Se guarda una orden
 * 2. Se guardan los productos de la orden y se asignan a un envio dependiendo del origen
 * 3. Se reduce el stock de cada producto segun la sucursal de donde seran enviados
 * 4. Se envia una notificacion general mediante Teams y notificaciones a las sucursales de donde deberan salir los productos
 * 5. Se envia un email para notificar en caso de haber sido ya pagado
 */
class Save{
    /**
     *
     *
     * @var Order
     */
    private $order;
    /**
     * Costo total de la orden
     *
     * @var float
     */
    private $total = 0;
    /**
     * Envio local
     *
     * @var Shipment
     */
    private $localShipment;
    /**
     * Envio nacional
     *
     * @var Shipment
     */
    private $nationalShipment;
    /**
     * Productos que seran enviados en la notificacion para la sucursal local
     *
     * @var array
     */
    private $productsForNotificationLocal = [];
    /**
     * Productos que seran enviados en la notificacion para la sucursal nacional
     *
     * @var array
     */
    private $productsForNotificationNational = [];
    /**
     * Indicativo de si la orden fue pagada o no
     *
     * @var [type]
     */
    private $payment_submitted;

    public function __construct(
        private Customer $customer,
        private array $requestData,
        private CartCollection $products,
        private Rules $rules)
    {

    }

    public function handle(){

        $this->calculateTotal();

        try {

            DB::beginTransaction();

            $this->createOrder();
            CartCondition::saveConditionsByOrder($this->order->id);
            if ($this->hasLocalLocation())
                $this->localShipment = ShipmentService::shipment(true, $this->order, $this->customer, $this->requestData, true);
            if ($this->hasNationalLocation())
                $this->nationalShipment = ShipmentService::shipment(false, $this->order, $this->customer, $this->requestData, false);

            $this->saveItems();

            if ($this->payment_submitted){
                Notification::notify(
                    $this->customer,
                    $this->order,
                    $this->productsForNotificationLocal,
                    $this->productsForNotificationNational,
                    $this->localShipment,
                    $this->nationalShipment
                );
            }

            DB::commit();

            return $this->order;

        } catch (\Throwable $th) {
            Handler::log($th);

            DB::rollBack();

            return null;
        }
    }
    /**
     * Calcular el total de la orden
     * 1. Se deben agregar los cargos y descuentos en caso de existir, posteriormente obtener el costo del los productos
     * 2. Agregar el costo del envio
     * @return void
     */
    private function calculateTotal() {
        $this->applyCartConditions();
        $this->addShippingCosts();
    }
    /**
     * Calcular el costo de total de los productos, aqui se aplican cargos y descuentos en caso de existir
     *
     * @return void
     */
    private function applyCartConditions() {
        CartConditions::chargeCartCondition($this->rules);
        $this->total = \Cart::getTotal();
    }
    /**
     * Costos del envio, en caso de existir
     *
     * @return void
     */
    private function addShippingCosts() {
        if ($this->hasLocalLocation()) {
            $total = ShipmentService::getTotal(
                        'local',
                        $this->getData('local_carrier'),
                        $this->getData('local_service_code')
                    );
            $this->total = Money::add($this->total, $total);
        }

        if ($this->hasNationalLocation()) {
            $total = ShipmentService::getTotal(
                'national',
                $this->getData('national_carrier'),
                $this->getData('national_service_code')
            );
            $this->total = Money::add($this->total, $total);
        }
    }

    private function hasLocalLocation(): bool {
        return !is_null($this->getData('local_warehouse_id'));
    }

    private function hasNationalLocation(): bool {
        return !is_null($this->getData('national_warehouse_id'));
    }
    private function getData($name){
        return key_exists($name, $this->requestData) ? $this->requestData[$name] : null;
    }
    private function createOrder(){

        $this->payment_submitted = $this->getData('payment_submitted');

        $this->order = Order::create([
            'customer_id' => $this->customer->id,
            'folio' => $this->folio(),
            'total' => $this->total,
            'payment_concept' => $this->getData('payment_concept'),
            'status' => $this->payment_submitted && $this->payment_submitted == 1  ? 'payment_submitted' : 'pending_payment',
        ]);

    }
    /**
     * Obtener el folio unico y consecutivo para el cliente
     *
     * @return int
     */
    private function folio(){
        $lastOrder = Order::where('customer_id', $this->customer->id)
            ->lockForUpdate()
            ->orderBy('folio', 'desc')
            ->first();

        return $lastOrder ? $lastOrder->folio + 1 : 1;
    }
    /**
     * Guardar los productos de la orden
     *
     * @return void
     */
    private function saveItems(){

        foreach ($this->products as $item) {
            $this->processItem($item);
            $this->updateStock($item);
        }
    }
    /**
     * Procesar un item, este puede ser local o nacional
     * @param [type] $item
     * @param array $localProductIds
     * @param array $nationalProductIds
     * @return void
     */
    private function processItem($item){
        $productNotification = new ObjectsProduct($item->name, $item->quantity, $item->price, $item->attributes->featured);

        if(is_numeric($item->attributes->location_id)){
            $this->productsForNotificationLocal[] = $productNotification;
            $this->createOrderItem($item, $this->localShipment->id);
        }else {
            $this->productsForNotificationNational[] = $productNotification;
            $this->createOrderItem($item, $this->nationalShipment->id);
        }
    }
    /**
     * Guardar un producto de una orden
     *
     * @param [type] $item
     * @param [type] $shipmentId
     * @return void
     */
    private function createOrderItem($item, $shipmentId){
        OrderItem::create([
            'order_id' => $this->order->id,
            'product_id' => $item->attributes->product_id,
            'quantity' => $item->quantity,
            'price' => $item->price,
            'shipment_id'=>  $shipmentId,
            'weight' => $item->attributes->weight,
            'length' => $item->attributes->length,
            'width' => $item->attributes->width,
            'height' => $item->attributes->height,
        ]);
    }
    /**
     * Actualizar el stock una vez que un producto de la orden es procesado
     *
     * @param [type] $item
     * @return void
     */
    private function updateStock($item){
        $locationId = $item->attributes->location == 'Nacional'
                        ? $this->getData('national_warehouse_id')
                        : $this->getData('local_warehouse_id');

        $productStock = ProductStock::where('product_id', $item->attributes->product_id)
            ->where('location_id', $locationId)
            ->first();

        if ($productStock) {
            $productStock->quantity_available -= $item->quantity;
            $productStock->quantity_on_order += $item->quantity;
            $productStock->save();
        } else {
            throw new \Exception('Stock not found for product: ' . $item->attributes->product_id);
        }
    }
}
