<?php

namespace App\UseCases\Order\Customer\SaveByRequest;

use App\Models\Customer;
use App\Models\Location;
use App\Models\Order;
use App\Models\Shipment;
use App\Services\ShippingServiceFactory;
use App\UseCases\Shipment\QuotationInCache;
use Illuminate\Support\Facades\Cache;
use App\Services\SelfPickup\Service as SelfPickupService;

class ShipmentService
{
    private const KEYS_IN_REQUEST = [
        'local' => [
            'shipping_cost' => 'local_shipping_cost',
            'destination_address_id' => 'local_destination_address_id',
            'warehouse_id' => 'local_warehouse_id',
            'carrier' => 'local_carrier',
            'service_code' => 'local_service_code'
        ],
        'national' =>[
            'destination_address_id' => 'national_destination_address_id',
            'warehouse_id' => 'national_warehouse_id',
            'carrier' => 'national_carrier',
            'shipping_cost' => 'national_shipping_cost',
            'service_code' => 'national_service_code'
        ]
    ];

    public static function shipment(bool $isLocal, Order $order, Customer $customer, array $requestData)
    {
        $data = self::getData($isLocal, $requestData);
        $quotationType = $isLocal ? 'local' : 'national';
        $customerAddress = $customer->addresses()->findOrFail($data['destination_address_id']);
        $warehouse = Location::findOrFail($data['warehouse_id']);

        return Shipment::create([
            'order_id' => $order->id,
            'customer_id' => $customer->id,
            'carrier' => $data['carrier'],
            'shipping_cost' => self::getTotal($quotationType, $data['carrier'], $data['service_code']),
            'is_local_shipping' => $isLocal,
            'service_code' => $data['service_code'],
            'origin_postal_code' => $warehouse->postal_code,
            'origin_address' => $warehouse->street,
            'origin_neighborhood' => $warehouse->neighborhood,
            'origin_state' => $warehouse->state,
            'origin_contact_phone' => $warehouse->contact_phone,
            'origin_city' => $warehouse->city,
            'origin_country' => $warehouse->country,
            'origin_number_ext' => $warehouse->number_ext,
            'origin_number_int' => $warehouse->number_int,
            'customer_address_id' => $customerAddress->id,
            'destination_postal_code' => $customerAddress->postal_code,
            'destination_address' => $customerAddress->street,
            'destination_neighborhood' => $customerAddress->neighborhood,
            'destination_state' => $customerAddress->state,
            'destination_contact_phone' => $customerAddress->contact_phone,
            'destination_city' => $customerAddress->city,
            'destination_country' => $customerAddress->country,
            'destination_number_ext' => $customerAddress->number_ext,
            'destination_number_int' => $customerAddress->number_int,
            'location_id' => $warehouse->id
        ]);
    }
    /**
     * El costo del servicio previamente cotizado, si el cliente es el carrier se provee un costo predeterminado
     *
     * @param string $type
     * @param string $carrier
     * @param string $service
     * @return string
     */
    public static function getTotal(string $type, string $carrier, string $service): string{

        if($carrier == ShippingServiceFactory::CUSTOMER_CARRIER) return SelfPickupService::defaultCost();

        $quotationInCache = new QuotationInCache(Cache::getStore());

        return $quotationInCache->get($type, $carrier, $service);
    }

    private static function getData(bool $isLocal, array $requestData): array
    {
        $keysInRequest = $isLocal ? self::KEYS_IN_REQUEST['local'] : self::KEYS_IN_REQUEST['national'];
        $data = [];

        foreach ($keysInRequest as $keyName => $keyInRequest) {
            if (key_exists($keyInRequest, $requestData))
                $data[$keyName] = $requestData[$keyInRequest];
            else
                throw new \Exception("Solicitud incompleta, parametro faltante: $keyInRequest", 1);
        }

        return $data;
    }
}
