<?php
namespace App\UseCases\PackagesSent\Viewer;

use App\Models\PackageSent;
use Illuminate\Support\Collection;

/**
 * Esta clase representa una vista tridimensional de los paquetes enviados usando la libreria threejs
 */
class Shipments{

    private $data = [];
    /**
     *
     *
     * @param array $packagesSent Paquetes enviados
     * @param array $itemsInPackages Productos dentro de los paquetes
     * @param Collection $shipmentIds Identificadores de los envios
     * @param Collection $productsInOrder Productos dentro de los envios, no es lo mismo un producto dentro de un paquete que dentro de una orden,
     * por ejemplo: puedes solicitar 4 paneles solares de un mismo tipo, lo que implica 4 productos dentro de lo paquetes
     * y un solo producto en la orden con una cantidad de 4
     */
    public function __construct(
        private readonly array $packagesSent,
        private readonly array $itemsInPackages,
        private readonly Collection $shipmentIds,
        private readonly Collection $productsInOrder)
    {

    }
    public function handle(): array{
        $this->data = [];

        foreach ($this->shipmentIds as $shipmentId) {

            $packagesOfShipment = $this->packagesByShipment($shipmentId);

            if($packagesOfShipment){
                foreach ($packagesOfShipment as $package) {
                    $packageView = $this->containerOfPackage($package, $shipmentId);

                    $productsByPackage = $this->productsByPackage($package->id);

                    if ($productsByPackage) {
                        $this->addProductsToContainer($packageView, $productsByPackage);
                    }

                    $this->data[] = $packageView;

                }
            }
        }

        return $this->data;
    }
    /**
     * Contenedor de los paquetes, pueden suceder dos situaciones, el paquete esta empacado o es autoembalaje
     *
     * @param PackageSent $package
     * @param integer $shipmentId
     * @return Container
     */
    private function containerOfPackage(PackageSent $package, int $shipmentId): Container{
        $containerView = new Container(
            $package->id,
            $shipmentId,
            $package->shipping_box_name ?? __('shipment.self-packaging')
        );
        if($package->shipping_box_id){
            $containerView->setGeometry(
                new BoxGeometry(
                    $package->shipping_box_outer_width,
                    $package->shipping_box_outer_height,
                    $package->shipping_box_outer_length
                ))
            ->setPosition(new Position(0, 0 , 0));
        }

        return $containerView;
    }
    /**
     * // Three.js utiliza un sistema de coordenadas donde el eje Y corresponde a la altura (arriba/abajo),
     * el eje X a la anchura (izquierda/derecha) y el eje Z a la profundidad (adelante/atrás).
     * Este esquema difiere de librerías como BoxPacker, que emplea un sistema en el que el eje X sigue representando la anchura,
     * pero el eje Y indica la profundidad y el eje Z la altura.
     *
     * Debido a esta inversión de los ejes Y y Z entre ambas librerías,
     * también cambia la interpretación de las dimensiones del objeto.
     * En BoxPacker, la altura se representa en el eje Z, mientras que en Three.js la altura se representa en el eje Y.
     * Por lo tanto, al convertir dimensiones entre las dos, es necesario intercambiar los valores de alto y profundidad.
     *
     * @param Container $containerView
     * @param Collection $productsPackaged
     * @return void
     */
    private function addProductsToContainer(Container $containerView, Collection $productsPackaged){
        foreach ($productsPackaged as $packagedProduct) {
            $product = $this->productsInOrder->firstWhere('id', $packagedProduct->order_item_id);

            if(!$product) throw new \Exception("No se encontro informacion del producto embalado", 1);

            $productView = new Item(
                $packagedProduct->id,
                $product['name'],
                new BoxGeometry(
                    $packagedProduct->width,
                    $packagedProduct->depth,
                    $packagedProduct->length
                ),
                new Position(
                    $packagedProduct->x,
                    $packagedProduct->z,
                    $packagedProduct->y
                ),
                $product
            );

            $containerView->add($productView);
        }
    }
    private function packagesByShipment(int $shipmentId): Collection|null{
        return key_exists($shipmentId, $this->packagesSent) ? $this->packagesSent[$shipmentId] : null;
    }
    private function productsByPackage(int $packageId): Collection|null{
        return key_exists($packageId, $this->itemsInPackages) ? $this->itemsInPackages[$packageId] : null;
    }
}
