<?php
namespace App\UseCases\PackagesSent\Viewer;

use Illuminate\Support\Collection;
use App\Services\Shipments\Packer;
use App\Services\Shipments\PerSend\RegularContainer;
use App\Services\Shipments\PerSend\SelfContainer;
use Ramsey\Uuid\Uuid;
use App\Services\Shipments\Contracts\Product;
use DVDoug\BoxPacker\PackedItem;
/**
 * Esta clase representa la representacion visual 3D de el embalaje propuesto.
 */
class ShipmentsPropose{

    private array $data = [];

    public function __construct(
        private readonly array $shipmentIds,
        private readonly Collection $productsInOrder
    )
    {

    }
    public function handle(): array{
        $this->data = [];

        foreach ($this->shipmentIds as $shipmentId) {

            $listPackaging = Packer::byShipment($shipmentId);
            $allPackages = $listPackaging->packages();
            foreach ($allPackages as $package) {
                $packageView = $this->containerOfPackage($package, $shipmentId);

                if($package instanceof RegularContainer){
                    $this->addProductsToContainer($packageView, $package);
                }

                $this->data[] = $packageView;

            }
        }

        return $this->data;
    }
    /**
     * Contenedor de los paquetes, pueden suceder dos situaciones, el paquete esta empacado o es autoembalaje
     *
     * @param PackageSent $package
     * @param integer $shipmentId
     * @return Container
     */
    private function containerOfPackage(RegularContainer|SelfContainer $package, int $shipmentId): Container{
        $uuid = Uuid::uuid4();

        if($package instanceof RegularContainer){
            $containerView = new Container(
                $uuid->toString(),
                $shipmentId,
                $package->shippingBox->name
            );
            $containerView->setGeometry(
                new BoxGeometry(
                    $package->shippingBox->outer_width,
                    $package->shippingBox->outer_height,
                    $package->shippingBox->outer_length
                ))
            ->setPosition(new Position(0, 0 , 0));
        }
        else if($package instanceof SelfContainer){
            $containerView = new Container(
                $uuid->toString(),
                $shipmentId,
                __('shipment.self-packaging')
            );
        }

        return $containerView;
    }
    /**
     * // Three.js utiliza un sistema de coordenadas donde el eje Y corresponde a la altura (arriba/abajo),
     * el eje X a la anchura (izquierda/derecha) y el eje Z a la profundidad (adelante/atrás).
     * Este esquema difiere de librerías como BoxPacker, que emplea un sistema en el que el eje X sigue representando la anchura,
     * pero el eje Y indica la profundidad y el eje Z la altura.
     *
     * Debido a esta inversión de los ejes Y y Z entre ambas librerías,
     * también cambia la interpretación de las dimensiones del objeto.
     * En BoxPacker, la altura se representa en el eje Z, mientras que en Three.js la altura se representa en el eje Y.
     * Por lo tanto, al convertir dimensiones entre las dos, es necesario intercambiar los valores de alto y profundidad.
     *
     * @param Container $containerView
     * @param RegularContainer $productsPackaged
     * @return void
     */
    private function addProductsToContainer(Container $containerView, RegularContainer $package){
        $packedItems = $package->packedBox->getItems();

        foreach ($packedItems as $packedItem) {
            $item = $packedItem->getItem();

            if($item instanceof Product && $packedItem instanceof PackedItem){
                $product = $this->productsInOrder->firstWhere('id', $item->identification());
                $uuid = Uuid::uuid4();

                if(!$product) throw new \Exception("No se encontro informacion del producto embalado", 1);

                $productView = new Item(
                    $uuid->toString(),
                    $product['name'],
                    new BoxGeometry(
                        $item->getWidth(),
                        $item->getDepth(),
                        $item->getLength()
                    ),
                    new Position(
                        $packedItem->getX(),
                        $packedItem->getZ(),
                        $packedItem->getY()
                    ),
                    $product
                );

                $containerView->add($productView);
            }
        }
    }
}
