<?php

namespace App\UseCases\Shipment;

use App\Exceptions\Handler;
use App\Models\Order;
use App\Models\Shipment;
use App\Models\ShipmentCarrierLog;
use App\Services\Contracts\ShippingServiceInterface;
use App\Services\Helpers\Money;
use Illuminate\Support\Facades\DB;
use App\Services\Contracts\Shipment as ContractsShipment;
/**
 * Actualizar la guia previamente solicitada para un envio
 * 1. Cambiar el costo de la orden
 * 2. Cambiar la informacion del envio
 * 3. Generar la guia en la paqueteria solicitada
 */
class UpdateGuide{
    /**
     *
     *
     * @param Order $order
     * @param Shipment $shipment
     * @param ShippingServiceInterface $shippingService
     * @param string $carrier
     * @param string $service_code
     * @param string $shipping_cost Debe venir ya en dolares
     * @param string $reason
     * @param integer $userId
     */
    public function __construct(
        private readonly Order $order,
        private readonly Shipment $shipment,
        private readonly ShippingServiceInterface $shippingService,
        private readonly string $carrier,
        private readonly string $service_code,
        private readonly string $shipping_cost,
        private readonly string $reason,
        private readonly int $userId)
    {

    }

    public function handle(): ContractsShipment|null{
        try {
            DB::beginTransaction();

            $this->updateCostOrder();
            $this->updateShipment();

            $result = $this->createGuide();

            if($result) $this->logShipmentUpdate();

            DB::commit();

            return $result;

        } catch (\Throwable $th) {
            DB::rollBack();
            Handler::log($th);
        }

        return null;
    }
    private function createGuide(){
        $handler = new Guide($this->shipment, $this->shippingService);
        return $handler->handle();
    }
    private function logShipmentUpdate(){
        ShipmentCarrierLog::create([
            'user_id' => $this->userId,
            'shipment_id' => $this->shipment->id,
            'reason' => $this->reason,
            'carrier' => $this->carrier,
            'service_code' => $this->service_code,
            'shipping_cost' => $this->shipping_cost
        ]);
    }
    private function updateShipment(){
        $this->shipment->carrier = $this->carrier;
        $this->shipment->service_code = $this->service_code;
        $this->shipment->shipping_cost = $this->shipping_cost;
        $this->shipment->save();
    }
    /**
     * Esto debe hacerse antes de actualizar el envio, necesitamos el costo del envio antes de ser actualizado
     *
     * @return void
     */
    private function updateCostOrder(){
        $shipmentOldCost = $this->shipment->shipping_cost;
        $cost = $this->order->total;

        $cost = Money::sub($cost, $shipmentOldCost);
        $cost = Money::add($cost, $this->shipping_cost);

        $this->order->total = $cost;
        $this->order->save();

    }
}
