import * as THREE from 'three';
import { OrbitControls } from 'three/addons/controls/OrbitControls.js';
import { BoxesColors } from './BoxesColors';
import { Tooltip } from './Tooltip';

export const Packaging = {
    packagesView: [],
    defaultPositionCamera: {
        x: 70,
        y: 30,
        z: 400
    },
    defaultBoxConfig: {
        color: '#FFBE10',
        wireframe: true,
        transparent: true,
    },
    defaultBoxPosition: {
        x: 0,
        y: 0,
        z: 0
    },
    defaultConfigCamera:{
        fov: 75,
        aspect: window.innerWidth / window.innerHeight,
        near: 0.1,
        far: 5000
    },
    init: function(packagesView){
        Packaging.packagesView = packagesView;
        Tooltip.init();
    },
    show: function(e){
        const packagingId = this.dataset.package;

        Swal.fire({
            html: `<div id="container-of-packaging" style="overflow: hidden;"></div>`,
            width: '99%',
            showCloseButton: true,
            showCancelButton: false,
            focusConfirm: true,
            confirmButtonText: 'Cerrar',
            didOpen: () => {
                const containerNode = document.getElementById("container-of-packaging");
                Packaging.createScene(packagingId, containerNode);
                document.addEventListener('pointermove', Tooltip.infoBox, false);
            },
            willClose: () => {
                Tooltip.close();
                document.removeEventListener('pointermove', Tooltip.infoBox, false);
            }
          })
    },
    createScene: function(packagingId, containerNode){
        const packageView = Packaging.packagesView.find((packageView) => packageView.id == packagingId);
        // Crear escena
        const scene = new THREE.Scene();
        // Crear cámara (perspectiva)
        const camera = Packaging.createCamera(packageView);
        // Crear renderizador
        const renderer = new THREE.WebGLRenderer();
        renderer.setSize(window.innerWidth -100, window.innerHeight - 100);
        containerNode.appendChild(renderer.domElement);
        //Controles
        const controls = new OrbitControls(camera, renderer.domElement);
        controls.update();
        // Crear el cubo contenedor
        const container = Packaging.containerBox(packageView)
        scene.add(container);
        // Crear los cubos para cada paquete
        Packaging.boxesShipment(container, packageView);

        Tooltip.setScene(scene, camera);

        const animate = () => {
            requestAnimationFrame(animate);

            renderer.render(scene, camera);
        }

        animate();
    },
    createCamera: function(packageView){
        const camera = new THREE.PerspectiveCamera(
            Packaging.defaultConfigCamera.fov,
            Packaging.defaultConfigCamera.aspect,
            Packaging.defaultConfigCamera.near,
            Packaging.defaultConfigCamera.far);

        camera.position.set(
            Packaging.defaultPositionCamera.x,
            packageView.boxGeometry.height,
            packageView.boxGeometry.depth );

        camera.lookAt(Packaging.defaultBoxPosition.x, Packaging.defaultBoxPosition.y, Packaging.defaultBoxPosition.z);

        return camera;
    },
    containerBox: function(packageView){
        // Crear geometría del contenedor
        const geometry = new THREE.BoxGeometry(
            packageView.boxGeometry.width,
            packageView.boxGeometry.height,
            packageView.boxGeometry.depth);
        // Material translúcido para el contenedor
        const material = new THREE.MeshBasicMaterial(Packaging.defaultBoxConfig);
        // Crear el cubo contenedor
        const container = new THREE.Mesh(geometry, material);
        container.position.set(0, 0, 0);
        container.userData = { isContainer: true };
        return container;
    },
    calcPositionX: function(packageView, positionX, itemWidth){
        return packageView.boxGeometry.width / 2 - positionX - itemWidth / 2
    },
    calcPositionY: function(packageView, positionY, height){
        return (packageView.boxGeometry.height / 2 - height / 2 - positionY) * -1
    },
    calcPositionZ: function(packageView, positionZ, deep){
        return packageView.boxGeometry.depth / 2 - deep / 2 - positionZ
    },
    boxShipment: function(item){
        const geometry = new THREE.BoxGeometry(
            item.boxGeometry.width,
            item.boxGeometry.height,
            item.boxGeometry.depth);
        const material = new THREE.MeshBasicMaterial(
            {
                color: BoxesColors.get() ,
                transparent: true,
                opacity: 0.5
            });

        return new THREE.Mesh(geometry, material);
    },
    boxesShipment: function(container, packageView){
        packageView.items.forEach(item => {
            const shipmentCube = new Packaging.boxShipment(item);

            Packaging.boxShipmentPosition(item, packageView, shipmentCube);
            shipmentCube.name = "box_shipment";
            shipmentCube.userData = {...item, isInnerBox: true};

            container.add(shipmentCube);
        });
    },
    /**
     * Debido a que el contenedor es inicializado en el centro de la escena, debemos tomar esto en cuenta cuando
     * colocamos los paquetes dentro de el, para que aparezcan desde la esquina derecha inferior
     * @param {*} item
     * @param {*} packageView
     * @param {*} shipmentCube
     */
    boxShipmentPosition: function(item, packageView, shipmentCube){
        shipmentCube.position.set(
            Packaging.calcPositionX(packageView, item.position.x, item.boxGeometry.width),
            Packaging.calcPositionY(packageView, item.position.y, item.boxGeometry.height),
            Packaging.calcPositionZ(packageView, item.position.z, item.boxGeometry.depth)
        );
    },
}
